#/bin/env python3

import copy
import os
import re
import warnings


def rewrite_data_passing_to_use_volumes(
    workflow: dict,
    volume: dict,
    path_prefix: str = 'artifact_data/',
) -> dict:
    """Converts Argo workflow that passes data using artifacts to workflow that
    passes data using a single multi-write volume.

    Implementation: Changes file passing system from passing Argo artifacts to passing parameters holding volumeMount subPaths.

    Limitations:
    * All artifact file names must be the same (e.g. "data"). E.g. "/tmp/outputs/my_out_1/data".
      Otherwise consumer component that can receive data from different producers won't be able to specify input file path since files from different upstreams will have different names.
    * Passing constant arguments to artifact inputs is not supported for now. Only references can be passed.

    Args:
        workflow: Workflow (dict) that needs to be converted.
        volume: Kubernetes spec (dict) of a volume that should be used for data passing. The volume should support multi-write (READ_WRITE_MANY) if the workflow has any parallel steps. Example: {'persistentVolumeClaim': {'claimName': 'my-pvc'}}

    Returns:
        Converted workflow (dict).
    """

    # Limitation: All artifact file names must be the same (e.g. "data"). E.g. "/tmp/outputs/my_out_1/data".
    # Otherwise consumer component that can receive data from different producers won't be able to specify input file path since files from different upstreams will have different names.
    # Maybe this limitation can be lifted, but it is not trivial. Maybe with DSL compile it's easier since the graph it creates is usually a tree. But recursion creates a real graph.
    # Limitation: Passing constant values to artifact inputs is not supported for now.

    # Idea: Passing volumeMount subPaths instead of artifcats

    workflow = copy.deepcopy(workflow)
    templates = workflow['spec']['templates']

    container_templates = [
        template for template in templates if 'container' in template
    ]
    dag_templates = [template for template in templates if 'dag' in template]
    steps_templates = [
        template for template in templates if 'steps' in template
    ]

    execution_data_dir = path_prefix + '{{workflow.uid}}_{{pod.name}}/'

    data_volume_name = 'data-storage'
    volume['name'] = data_volume_name

    subpath_parameter_name_suffix = '-subpath'

    def convert_artifact_reference_to_parameter_reference(
            reference: str) -> str:
        parameter_reference = re.sub(
            r'{{([^}]+)\.artifacts\.([^}]+)}}',
            r'{{\1.parameters.\2' + subpath_parameter_name_suffix +
            '}}',  # re.escape(subpath_parameter_name_suffix) escapes too much.
            reference,
        )
        return parameter_reference

    # Adding the data storage volume to the workflow
    workflow['spec'].setdefault('volumes', []).append(volume)

    all_artifact_file_names = set(
    )  # All artifacts should have same file name (usually, "data"). This variable holds all different artifact names for verification.

    # Rewriting container templates
    for template in templates:
        if 'container' not in template and 'script' not in template:
            continue
        container_spec = template['container'] or template['steps']
        # Inputs
        input_artifacts = template.get('inputs', {}).get('artifacts', [])
        if input_artifacts:
            input_parameters = template.setdefault('inputs', {}).setdefault(
                'parameters', [])
            volume_mounts = container_spec.setdefault('volumeMounts', [])
            for input_artifact in input_artifacts:
                subpath_parameter_name = input_artifact[
                    'name'] + subpath_parameter_name_suffix  # TODO: Maybe handle clashing names.
                artifact_file_name = os.path.basename(input_artifact['path'])
                all_artifact_file_names.add(artifact_file_name)
                artifact_dir = os.path.dirname(input_artifact['path'])
                volume_mounts.append({
                    'mountPath':
                        artifact_dir,
                    'name':
                        data_volume_name,
                    'subPath':
                        '{{inputs.parameters.' + subpath_parameter_name + '}}',
                    'readOnly':
                        True,
                })
                input_parameters.append({
                    'name': subpath_parameter_name,
                })
        template.get('inputs', {}).pop('artifacts', None)

        # Outputs
        output_artifacts = template.get('outputs', {}).get('artifacts', [])
        if output_artifacts:
            output_parameters = template.setdefault('outputs', {}).setdefault(
                'parameters', [])
            del template.get('outputs', {})['artifacts']
            volume_mounts = container_spec.setdefault('volumeMounts', [])
            for output_artifact in output_artifacts:
                output_name = output_artifact['name']
                subpath_parameter_name = output_name + subpath_parameter_name_suffix  # TODO: Maybe handle clashing names.
                artifact_file_name = os.path.basename(output_artifact['path'])
                all_artifact_file_names.add(artifact_file_name)
                artifact_dir = os.path.dirname(output_artifact['path'])
                output_subpath = execution_data_dir + output_name
                volume_mounts.append({
                    'mountPath': artifact_dir,
                    'name': data_volume_name,
                    'subPath':
                        output_subpath,  # TODO: Switch to subPathExpr when it's out of beta: https://kubernetes.io/docs/concepts/storage/volumes/#using-subpath-with-expanded-environment-variables
                })
                output_parameters.append({
                    'name': subpath_parameter_name,
                    'value': output_subpath,  # Requires Argo 2.3.0+
                })
            whitelist = ['mlpipeline-ui-metadata', 'mlpipeline-metrics']
            output_artifacts = [
                artifact for artifact in output_artifacts
                if artifact['name'] in whitelist
            ]
            if not output_artifacts:
                template.get('outputs', {}).pop('artifacts', None)
            else:
                template.get('outputs',
                             {}).update({'artifacts': output_artifacts})

    # Rewrite DAG templates
    for template in templates:
        if 'dag' not in template and 'steps' not in template:
            continue

        # Inputs
        input_artifacts = template.get('inputs', {}).get('artifacts', [])
        if input_artifacts:
            input_parameters = template.setdefault('inputs', {}).setdefault(
                'parameters', [])
            volume_mounts = container_spec.setdefault('volumeMounts', [])
            for input_artifact in input_artifacts:
                subpath_parameter_name = input_artifact[
                    'name'] + subpath_parameter_name_suffix  # TODO: Maybe handle clashing names.
                input_parameters.append({
                    'name': subpath_parameter_name,
                })
        template.get('inputs', {}).pop('artifacts', None)

        # Outputs
        output_artifacts = template.get('outputs', {}).get('artifacts', [])
        if output_artifacts:
            output_parameters = template.setdefault('outputs', {}).setdefault(
                'parameters', [])
            volume_mounts = container_spec.setdefault('volumeMounts', [])
            for output_artifact in output_artifacts:
                output_name = output_artifact['name']
                subpath_parameter_name = output_name + subpath_parameter_name_suffix  # TODO: Maybe handle clashing names.
                output_parameters.append({
                    'name': subpath_parameter_name,
                    'valueFrom': {
                        'parameter':
                            convert_artifact_reference_to_parameter_reference(
                                output_artifact['from'])
                    },
                })
        template.get('outputs', {}).pop('artifacts', None)

        # Arguments
        for task in template.get('dag', {}).get('tasks', []) + [
                steps for group in template.get('steps', []) for steps in group
        ]:
            argument_artifacts = task.get('arguments', {}).get('artifacts', [])
            if argument_artifacts:
                argument_parameters = task.setdefault('arguments',
                                                      {}).setdefault(
                                                          'parameters', [])
                for argument_artifact in argument_artifacts:
                    if 'from' not in argument_artifact:
                        raise NotImplementedError(
                            'Volume-based data passing rewriter does not support constant artifact arguments at this moment. Only references can be passed.'
                        )
                    subpath_parameter_name = argument_artifact[
                        'name'] + subpath_parameter_name_suffix  # TODO: Maybe handle clashing names.
                    argument_parameters.append({
                        'name':
                            subpath_parameter_name,
                        'value':
                            convert_artifact_reference_to_parameter_reference(
                                argument_artifact['from']),
                    })
            task.get('arguments', {}).pop('artifacts', None)

        # There should not be any artifact references in any other part of DAG template (only parameter references)

    # Check that all artifacts have the same file names
    if len(all_artifact_file_names) > 1:
        warnings.warn(
            'Detected different artifact file names: [{}]. The workflow can fail at runtime. Please use the same file name (e.g. "data") for all artifacts.'
            .format(', '.join(all_artifact_file_names)))

    # Fail if workflow has argument artifacts
    workflow_argument_artifacts = workflow['spec'].get('arguments',
                                                       {}).get('artifacts', [])
    if workflow_argument_artifacts:
        raise NotImplementedError(
            'Volume-based data passing rewriter does not support constant artifact arguments at this moment. Only references can be passed.'
        )

    return workflow


if __name__ == '__main__':
    """Converts Argo workflow that passes data using artifacts to workflow that
    passes data using a single multi-write volume."""

    import argparse
    import io
    import yaml

    parser = argparse.ArgumentParser(
        prog='data_passing_using_volume',
        epilog='''Example: data_passing_using_volume.py --input argo/examples/artifact-passing.yaml --output argo/examples/artifact-passing-volumes.yaml --volume "persistentVolumeClaim: {claimName: my-pvc}"'''
    )
    parser.add_argument(
        "--input",
        type=argparse.FileType('r'),
        required=True,
        help='Path to workflow that needs to be converted.')
    parser.add_argument(
        "--output",
        type=argparse.FileType('w'),
        required=True,
        help='Path where to write converted workflow.')
    parser.add_argument(
        "--volume",
        type=str,
        required=True,
        help='''Kubernetes spec (YAML) of a volume that should be used for data passing. The volume should support multi-write (READ_WRITE_MANY) if the workflow has any parallel steps. Example: "persistentVolumeClaim: {claimName: my-pvc}".'''
    )
    args = parser.parse_args()

    input_workflow = yaml.safe_load(args.input)
    volume = yaml.safe_load(io.StringIO(args.volume))
    output_workflow = rewrite_data_passing_to_use_volumes(
        input_workflow, volume)
    yaml.dump(output_workflow, args.output)



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The SDK client for Kubeflow Pipelines API."""

import copy
import dataclasses
import datetime
import json
import logging
import os
import re
import tarfile
import tempfile
import time
from types import ModuleType
from typing import Any, Dict, List, Optional, TextIO
import warnings
import zipfile

from google.protobuf import json_format
from kfp import compiler
from kfp.client import auth
from kfp.client import set_volume_credentials
from kfp.components import base_component
from kfp.pipeline_spec import pipeline_spec_pb2
import kfp_server_api
import yaml

# Operators on scalar values. Only applies to one of |int_value|,
# |long_value|, |string_value| or |timestamp_value|.
_FILTER_OPERATIONS = {
    'EQUALS': 1,
    'NOT_EQUALS': 2,
    'GREATER_THAN': 3,
    'GREATER_THAN_EQUALS': 5,
    'LESS_THAN': 6,
    'LESS_THAN_EQUALS': 7,
    'IN': 8,
    'IS_SUBSTRING': 9,
}

KF_PIPELINES_ENDPOINT_ENV = 'KF_PIPELINES_ENDPOINT'
KF_PIPELINES_UI_ENDPOINT_ENV = 'KF_PIPELINES_UI_ENDPOINT'
KF_PIPELINES_DEFAULT_EXPERIMENT_NAME = 'KF_PIPELINES_DEFAULT_EXPERIMENT_NAME'
KF_PIPELINES_OVERRIDE_EXPERIMENT_NAME = 'KF_PIPELINES_OVERRIDE_EXPERIMENT_NAME'
KF_PIPELINES_IAP_OAUTH2_CLIENT_ID_ENV = 'KF_PIPELINES_IAP_OAUTH2_CLIENT_ID'
KF_PIPELINES_APP_OAUTH2_CLIENT_ID_ENV = 'KF_PIPELINES_APP_OAUTH2_CLIENT_ID'
KF_PIPELINES_APP_OAUTH2_CLIENT_SECRET_ENV = 'KF_PIPELINES_APP_OAUTH2_CLIENT_SECRET'


@dataclasses.dataclass
class _PipelineDoc:
    pipeline_spec: pipeline_spec_pb2.PipelineSpec
    platform_spec: pipeline_spec_pb2.PlatformSpec

    def to_dict(self) -> dict:
        if self.platform_spec == pipeline_spec_pb2.PlatformSpec():
            return json_format.MessageToDict(self.pipeline_spec)
        else:
            return {
                'pipeline_spec': json_format.MessageToDict(self.pipeline_spec),
                'platform_spec': json_format.MessageToDict(self.platform_spec),
            }


@dataclasses.dataclass
class _JobConfig:
    pipeline_spec: dict
    pipeline_version_reference: kfp_server_api.V2beta1PipelineVersionReference
    runtime_config: kfp_server_api.V2beta1RuntimeConfig


class RunPipelineResult:

    def __init__(self, client: 'Client',
                 run_info: kfp_server_api.V2beta1Run) -> None:
        self._client = client
        self.run_info = run_info
        self.run_id = run_info.run_id

    def wait_for_run_completion(self, timeout=None):
        timeout = timeout or datetime.timedelta.max
        return self._client.wait_for_run_completion(self.run_id, timeout)

    def __repr__(self):
        return f'RunPipelineResult(run_id={self.run_id})'


class Client:
    """The KFP SDK client for the Kubeflow Pipelines backend API.

    Args:
        host: Host name to use to talk to Kubeflow Pipelines. If not set,
            the in-cluster service DNS name will be used, which only works if
            the current environment is a pod in the same cluster (such as a
            Jupyter instance spawned by Kubeflow's JupyterHub). (`More information on connecting. <https://www.kubeflow.org/docs/components/pipelines/sdk/connect-api/>`_)
        client_id: Client ID used by Identity-Aware Proxy.
        namespace: Kubernetes namespace to use. Used for multi-user deployments. For single-user deployments, this should be left as ``None``.
        other_client_id: Client ID used to obtain the auth codes and refresh
            tokens (`reference <https://cloud.google.com/iap/docs/authentication-howto#authenticating_from_a_desktop_app>`_).
        other_client_secret: Client secret used to obtain the auth codes and
            refresh tokens.
        existing_token: Authentication token to pass in directly. Used in cases where the token is
            generated from outside the SDK.
        cookies: CookieJar object containing cookies that will be passed to the
            Pipelines API.
        proxy: HTTP or HTTPS proxy server.
        ssl_ca_cert: Certification for proxy.
        kube_context: kubectl context to use. Must be a context listed in the kubeconfig file. Defaults to the current-context set within kubeconfig.
        credentials: ``TokenCredentialsBase`` object which provides the logic to
            populate the requests with credentials to authenticate against the
            API server.
        ui_host: Base URL to use to open the Kubeflow Pipelines UI. This is used
            when running the client from a notebook to generate and print links.
        verify_ssl: Whether to verify the server's TLS certificate.
    """

    # in-cluster DNS name of the pipeline service
    _IN_CLUSTER_DNS_NAME = 'ml-pipeline.{}.svc.cluster.local:8888'
    _KUBE_PROXY_PATH = 'api/v1/namespaces/{}/services/ml-pipeline:http/proxy/'

    # Auto populated path in pods
    # https://kubernetes.io/docs/tasks/access-application-cluster/access-cluster/#accessing-the-api-from-a-pod
    # https://kubernetes.io/docs/reference/access-authn-authz/service-accounts-admin/#serviceaccount-admission-controller
    _NAMESPACE_PATH = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'

    _LOCAL_KFP_CONTEXT = os.path.expanduser('~/.config/kfp/context.json')

    # TODO: Wrap the configurations for different authentication methods.
    def __init__(
        self,
        host: Optional[str] = None,
        client_id: Optional[str] = None,
        namespace: str = 'kubeflow',
        other_client_id: Optional[str] = None,
        other_client_secret: Optional[str] = None,
        existing_token: Optional[str] = None,
        cookies: Optional[str] = None,
        proxy: Optional[str] = None,
        ssl_ca_cert: Optional[str] = None,
        kube_context: Optional[str] = None,
        credentials: Optional[str] = None,
        ui_host: Optional[str] = None,
        verify_ssl: Optional[bool] = None,
    ) -> None:
        """Create a new instance of kfp client."""
        warnings.warn(
            'This client only works with Kubeflow Pipeline v2.0.0-beta.2 '
            'and later versions.',
            category=FutureWarning)

        host = host or os.environ.get(KF_PIPELINES_ENDPOINT_ENV)
        self._uihost = os.environ.get(KF_PIPELINES_UI_ENDPOINT_ENV, ui_host or
                                      host)
        client_id = client_id or os.environ.get(
            KF_PIPELINES_IAP_OAUTH2_CLIENT_ID_ENV)
        other_client_id = other_client_id or os.environ.get(
            KF_PIPELINES_APP_OAUTH2_CLIENT_ID_ENV)
        other_client_secret = other_client_secret or os.environ.get(
            KF_PIPELINES_APP_OAUTH2_CLIENT_SECRET_ENV)
        config = self._load_config(host, client_id, namespace, other_client_id,
                                   other_client_secret, existing_token, proxy,
                                   ssl_ca_cert, kube_context, credentials,
                                   verify_ssl)
        # Save the loaded API client configuration, as a reference if update is
        # needed.
        self._load_context_setting_or_default()

        # If custom namespace provided, overwrite the loaded or default one in
        # context settings for current client instance
        if namespace != 'kubeflow':
            self._context_setting['namespace'] = namespace

        self._existing_config = config
        if cookies is None:
            cookies = self._context_setting.get('client_authentication_cookie')
        api_client = kfp_server_api.ApiClient(
            config,
            cookie=cookies,
            header_name=self._context_setting.get(
                'client_authentication_header_name'),
            header_value=self._context_setting.get(
                'client_authentication_header_value'))
        _add_generated_apis(self, kfp_server_api, api_client)
        self._recurring_run_api = kfp_server_api.RecurringRunServiceApi(
            api_client)
        self._run_api = kfp_server_api.RunServiceApi(api_client)
        self._experiment_api = kfp_server_api.ExperimentServiceApi(api_client)
        self._pipelines_api = kfp_server_api.PipelineServiceApi(api_client)
        self._upload_api = kfp_server_api.PipelineUploadServiceApi(api_client)
        self._healthz_api = kfp_server_api.HealthzServiceApi(api_client)
        if not self._context_setting['namespace'] and self.get_kfp_healthz(
        ).multi_user is True:
            try:
                with open(Client._NAMESPACE_PATH, 'r') as f:
                    current_namespace = f.read()
                    self.set_user_namespace(current_namespace)
            except FileNotFoundError:
                logging.info(
                    'Failed to automatically set namespace.', exc_info=False)

    def _load_config(
        self,
        host: Optional[str],
        client_id: Optional[str],
        namespace: str,
        other_client_id: Optional[str],
        other_client_secret: Optional[str],
        existing_token: Optional[str],
        proxy: Optional[str],
        ssl_ca_cert: Optional[str],
        kube_context: Optional[str],
        credentials: Optional[str],
        verify_ssl: Optional[bool],
    ) -> kfp_server_api.Configuration:
        config = kfp_server_api.Configuration()

        if proxy:
            # https://github.com/kubeflow/pipelines/blob/c6ac5e0b1fd991e19e96419f0f508ec0a4217c29/backend/api/python_http_client/kfp_server_api/rest.py#L100
            config.proxy = proxy
        if verify_ssl is not None:
            config.verify_ssl = verify_ssl

        if ssl_ca_cert:
            config.ssl_ca_cert = ssl_ca_cert

        host = host or ''

        # Defaults to 'https' if host does not contain 'http' or 'https' protocol.
        if host and not host.startswith('http'):
            warnings.warn(
                f'The host {host} does not contain the "http" or "https" protocol. Defaults to "https".'
            )
            host = 'https://' + host

        # Preprocess the host endpoint to prevent some common user mistakes.
        if not client_id:
            # always preserving the protocol (http://localhost requires it)
            host = host.rstrip('/')

        if host:
            config.host = host

        token = None

        # "existing_token" is designed to accept token generated outside of SDK.
        #
        # https://cloud.google.com/functions/docs/securing/function-identity
        # https://cloud.google.com/endpoints/docs/grpc/service-account-authentication
        #
        # Here is an example.
        #
        # import requests
        # import kfp
        #
        # def get_access_token():
        #     url = 'http://metadata.google.internal/computeMetadata/v1/instance/service-accounts/default/token'
        #     r = requests.get(url, headers={'Metadata-Flavor': 'Google'})
        #     r.raise_for_status()
        #     access_token = r.json()['access_token']
        #     return access_token
        #
        # client = kfp.Client(host='<KFPHost>', existing_token=get_access_token())
        #
        if existing_token:
            token = existing_token
            self._is_refresh_token = False
        elif client_id:
            token, self._is_refresh_token = auth.get_auth_token(
                client_id, other_client_id, other_client_secret)
        elif self._is_inverse_proxy_host(host):
            token = auth.get_gcp_access_token()
            self._is_refresh_token = False
        elif credentials:
            config.api_key['authorization'] = 'placeholder'
            config.api_key_prefix['authorization'] = 'Bearer'
            config.refresh_api_key_hook = credentials.refresh_api_key_hook

        if token:
            config.api_key['authorization'] = token
            config.api_key_prefix['authorization'] = 'Bearer'
            return config

        if host:
            # if host is explicitly set with auth token, it's probably a port
            # forward address.
            return config

        import kubernetes as k8s
        in_cluster = True
        try:
            k8s.config.load_incluster_config()
        except:
            in_cluster = False

        if in_cluster:
            config.host = Client._IN_CLUSTER_DNS_NAME.format(namespace)
            config = self._get_config_with_default_credentials(config)
            return config

        try:
            k8s.config.load_kube_config(
                client_configuration=config, context=kube_context)
        except:
            print('Failed to load kube config.')
            return config

        if config.host:
            config.host = config.host + '/' + Client._KUBE_PROXY_PATH.format(
                namespace)
        return config

    def _is_inverse_proxy_host(self, host: str) -> bool:
        return bool(re.match(r'\S+.googleusercontent.com/{0,1}$', host))

    def _is_ipython(self) -> bool:
        """Returns whether we are running in notebook."""
        try:
            import IPython
            ipy = IPython.get_ipython()
            if ipy is None:
                return False
        except ImportError:
            return False

        return True

    def _get_url_prefix(self) -> str:
        if self._uihost:
            # User's own connection.
            if self._uihost.startswith('http://') or self._uihost.startswith(
                    'https://'):
                return self._uihost
            else:
                return 'http://' + self._uihost

        # In-cluster pod. We could use relative URL.
        return '/pipeline'

    def _load_context_setting_or_default(self) -> None:
        if os.path.exists(Client._LOCAL_KFP_CONTEXT):
            with open(Client._LOCAL_KFP_CONTEXT, 'r') as f:
                self._context_setting = json.load(f)
        else:
            self._context_setting = {
                'namespace': '',
            }

    def _refresh_api_client_token(self) -> None:
        """Refreshes the existing token associated with the kfp_api_client."""
        if getattr(self, '_is_refresh_token', None):
            return

        new_token = auth.get_gcp_access_token()
        self._existing_config.api_key['authorization'] = new_token

    def _get_config_with_default_credentials(
            self, config: kfp_server_api.Configuration
    ) -> kfp_server_api.Configuration:
        """Apply default credentials to the configuration object.

        This method accepts a Configuration object and extends it with
        some default credentials interface.
        """
        # XXX: The default credentials are audience-based service account tokens
        # projected by the kubelet (ServiceAccountTokenVolumeCredentials). As we
        # implement more and more credentials, we can have some heuristic and
        # choose from a number of options.
        # See https://github.com/kubeflow/pipelines/pull/5287#issuecomment-805654121
        credentials = set_volume_credentials.ServiceAccountTokenVolumeCredentials(
        )
        config_copy = copy.deepcopy(config)

        try:
            credentials.refresh_api_key_hook(config_copy)
        except Exception:
            logging.warning('Failed to set up default credentials. Proceeding'
                            ' without credentials...')
            return config

        config.refresh_api_key_hook = credentials.refresh_api_key_hook
        config.api_key_prefix['authorization'] = 'Bearer'
        config.refresh_api_key_hook(config)
        return config

    def set_user_namespace(self, namespace: str) -> None:
        """Sets the namespace in the Kuberenetes cluster to use.

        This function should only be used when Kubeflow Pipelines is in the
        multi-user mode.

        Args:
            namespace: Namespace to use within the Kubernetes cluster (namespace containing the Kubeflow Pipelines deployment).
        """
        self._context_setting['namespace'] = namespace
        if not os.path.exists(os.path.dirname(Client._LOCAL_KFP_CONTEXT)):
            os.makedirs(os.path.dirname(Client._LOCAL_KFP_CONTEXT))
        with open(Client._LOCAL_KFP_CONTEXT, 'w') as f:
            json.dump(self._context_setting, f)

    def get_kfp_healthz(
        self,
        sleep_duration: int = 5,
    ) -> kfp_server_api.V2beta1GetHealthzResponse:
        """Gets healthz info for KFP deployment.

        Args:
            sleep_duration: Time in seconds between retries.

        Returns:
            JSON response from the healthz endpoint.
        """
        count = 0
        response = None
        max_attempts = 5
        while not response:
            count += 1
            if count > max_attempts:
                raise TimeoutError(
                    f'Failed getting healthz endpoint after {max_attempts} attempts.'
                )

            try:
                return self._healthz_api.get_healthz()
            # ApiException, including network errors, is the only type that may
            # recover after retry.
            except kfp_server_api.ApiException:
                # logging.exception also logs detailed info about the ApiException
                logging.exception(
                    f'Failed to get healthz info attempt {count} of {max_attempts}.'
                )
                time.sleep(sleep_duration)

    def get_user_namespace(self) -> str:
        """Gets user namespace in context config.

        Returns:
            Kubernetes namespace from the local context file or empty if it
            wasn't set.
        """
        return self._context_setting['namespace']

    def create_experiment(
        self,
        name: str,
        description: str = None,
        namespace: str = None,
    ) -> kfp_server_api.V2beta1Experiment:
        """Creates a new experiment.

        Args:
            name: Name of the experiment.
            description: Description of the experiment.
            namespace: Kubernetes namespace to use. Used for multi-user deployments. For single-user deployments, this should be left as ``None``.

        Returns:
            ``V2beta1Experiment`` object.
        """
        namespace = namespace or self.get_user_namespace()
        experiment = None
        try:
            experiment = self.get_experiment(
                experiment_name=name, namespace=namespace)
        except ValueError as error:
            # Ignore error if the experiment does not exist.
            if not str(error).startswith('No experiment is found with name'):
                raise error

        if not experiment:
            logging.info(f'Creating experiment {name}.')

            experiment = kfp_server_api.V2beta1Experiment(
                display_name=name,
                description=description,
                namespace=namespace,
            )
            experiment = self._experiment_api.create_experiment(body=experiment)

        link = f'{self._get_url_prefix()}/#/experiments/details/{experiment.experiment_id}'
        if self._is_ipython():
            import IPython
            html = f'<a href="{link}" target="_blank" >Experiment details</a>.'
            IPython.display.display(IPython.display.HTML(html))
        else:
            print(f'Experiment details: {link}')

        return experiment

    def get_pipeline_id(self, name: str) -> Optional[str]:
        """Gets the ID of a pipeline by its name.

        Args:
            name: Pipeline name.

        Returns:
            The pipeline ID if a pipeline with the name exists.
        """
        pipeline_filter = json.dumps({
            'predicates': [{
                'operation': _FILTER_OPERATIONS['EQUALS'],
                'key': 'display_name',
                'stringValue': name,
            }]
        })
        result = self._pipelines_api.list_pipelines(filter=pipeline_filter)
        if result.pipelines is None:
            return None
        if len(result.pipelines) == 1:
            return result.pipelines[0].pipeline_id
        elif len(result.pipelines) > 1:
            raise ValueError(
                f'Multiple pipelines with the name: {name} found, the name needs to be unique.'
            )
        return None

    def list_experiments(
        self,
        page_token: str = '',
        page_size: int = 10,
        sort_by: str = '',
        namespace: Optional[str] = None,
        filter: Optional[str] = None,
    ) -> kfp_server_api.V2beta1ListExperimentsResponse:
        """Lists experiments.

        Args:
            page_token: Page token for obtaining page from paginated response.
            page_size: Size of the page.
            sort_by: Sort string of format ``'[field_name]', '[field_name] desc'``. For example, ``'display_name desc'``.
            namespace: Kubernetes namespace to use. Used for multi-user deployments. For single-user deployments, this should be left as ``None``.
            filter: A url-encoded, JSON-serialized Filter protocol buffer
                (see `filter.proto message <https://github.com/kubeflow/pipelines/blob/cb7d9a87c999eb1d2280959e5afbeee9e270ef3d/backend/api/v2beta1/filter.proto>`_). Example:

                  ::

                    json.dumps({
                        "predicates": [{
                            "operation": "EQUALS",
                            "key": "display_name",
                            "stringValue": "my-name",
                        }]
                    })

        Returns:
            ``V2beta1ListExperimentsResponse`` object.
        """
        namespace = namespace or self.get_user_namespace()
        return self._experiment_api.list_experiments(
            page_token=page_token,
            page_size=page_size,
            sort_by=sort_by,
            filter=filter,
            namespace=namespace,
        )

    def get_experiment(
        self,
        experiment_id: Optional[str] = None,
        experiment_name: Optional[str] = None,
        namespace: Optional[str] = None,
    ) -> kfp_server_api.V2beta1Experiment:
        """Gets details of an experiment.

        Either ``experiment_id`` or ``experiment_name`` is required.

        Args:
            experiment_id: ID of the experiment.
            experiment_name: Name of the experiment.
            namespace: Kubernetes namespace to use. Used for multi-user deployments.
                For single-user deployments, this should be left as ``None``.

        Returns:
            ``V2beta1Experiment`` object.
        """
        namespace = namespace or self.get_user_namespace()
        if experiment_id is None and experiment_name is None:
            raise ValueError(
                'Either experiment_id or experiment_name is required.')
        if experiment_id is not None:
            return self._experiment_api.get_experiment(
                experiment_id=experiment_id)
        experiment_filter = json.dumps({
            'predicates': [{
                'operation': _FILTER_OPERATIONS['EQUALS'],
                'key': 'display_name',
                'stringValue': experiment_name,
            }]
        })
        if namespace is not None:
            result = self._experiment_api.list_experiments(
                filter=experiment_filter, namespace=namespace)
        else:
            result = self._experiment_api.list_experiments(
                filter=experiment_filter)
        if not result.experiments:
            raise ValueError(
                f'No experiment is found with name {experiment_name}.')
        if len(result.experiments) > 1:
            raise ValueError(
                f'Multiple experiments is found with name {experiment_name}.')
        return result.experiments[0]

    def archive_experiment(self, experiment_id: str) -> dict:
        """Archives an experiment.

        Args:
            experiment_id: ID of the experiment.

        Returns:
            Empty dictionary.
        """
        return self._experiment_api.archive_experiment(
            experiment_id=experiment_id)

    def unarchive_experiment(self, experiment_id: str) -> dict:
        """Unarchives an experiment.

        Args:
            experiment_id: ID of the experiment.

        Returns:
            Empty dictionary.
        """
        return self._experiment_api.unarchive_experiment(
            experiment_id=experiment_id)

    def delete_experiment(self, experiment_id: str) -> dict:
        """Delete experiment.

        Args:
            experiment_id: ID of the experiment.

        Returns:
            Empty dictionary.
        """
        return self._experiment_api.delete_experiment(
            experiment_id=experiment_id)

    def list_pipelines(
        self,
        page_token: str = '',
        page_size: int = 10,
        sort_by: str = '',
        filter: Optional[str] = None,
        namespace: Optional[str] = None,
    ) -> kfp_server_api.V2beta1ListPipelinesResponse:
        """Lists pipelines.

        Args:
            page_token: Page token for obtaining page from paginated response.
            page_size: Size of the page.
            sort_by: Sort string of format ``'[field_name]', '[field_name] desc'``. For example, ``'display_name desc'``.
            filter: A url-encoded, JSON-serialized Filter protocol buffer
                (see `filter.proto message <https://github.com/kubeflow/pipelines/blob/cb7d9a87c999eb1d2280959e5afbeee9e270ef3d/backend/api/v2beta1/filter.proto>`_). Example:

                  ::

                    json.dumps({
                        "predicates": [{
                            "operation": "EQUALS",
                            "key": "display_name",
                            "stringValue": "my-name",
                        }]
                    })

        Returns:
            ``V2beta1ListPipelinesResponse`` object.
        """
        return self._pipelines_api.list_pipelines(
            namespace=namespace,
            page_token=page_token,
            page_size=page_size,
            sort_by=sort_by,
            filter=filter)

    # TODO: provide default namespace, similar to kubectl default namespaces.
    def run_pipeline(
        self,
        experiment_id: str,
        job_name: str,
        pipeline_package_path: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
        pipeline_id: Optional[str] = None,
        version_id: Optional[str] = None,
        pipeline_root: Optional[str] = None,
        enable_caching: Optional[bool] = None,
        service_account: Optional[str] = None,
    ) -> kfp_server_api.V2beta1Run:
        """Runs a specified pipeline.

        Args:
            experiment_id: ID of an experiment.
            job_name: Name of the job.
            pipeline_package_path: Local path of the pipeline package (the
                filename should end with one of the following .tar.gz, .tgz,
                .zip, .json).
            params: Arguments to the pipeline function provided as a dict.
            pipeline_id: ID of the pipeline.
            version_id: ID of the pipeline version to run.
                If both pipeline_id and version_id are specified, version_id
                will take precendence.
                If only pipeline_id is specified, the default version of this
                pipeline is used to create the run.
            pipeline_root: Root path of the pipeline outputs.
            enable_caching: Whether or not to enable caching for the
                run. If not set, defaults to the compile-time settings, which
                is ``True`` for all tasks by default. If set, the
                setting applies to all tasks in the pipeline (overrides the
                compile time settings).
            service_account: Specifies which Kubernetes service
                account to use for this run.

        Returns:
            ``V2beta1Run`` object.
        """
        job_config = self._create_job_config(
            params=params,
            pipeline_package_path=pipeline_package_path,
            pipeline_id=pipeline_id,
            version_id=version_id,
            enable_caching=enable_caching,
            pipeline_root=pipeline_root,
        )

        run_body = kfp_server_api.V2beta1Run(
            experiment_id=experiment_id,
            display_name=job_name,
            pipeline_spec=job_config.pipeline_spec,
            pipeline_version_reference=job_config.pipeline_version_reference,
            runtime_config=job_config.runtime_config,
            service_account=service_account)

        response = self._run_api.create_run(body=run_body)

        link = f'{self._get_url_prefix()}/#/runs/details/{response.run_id}'
        if self._is_ipython():
            import IPython
            html = (f'<a href="{link}" target="_blank" >Run details</a>.')
            IPython.display.display(IPython.display.HTML(html))
        else:
            print(f'Run details: {link}')

        return response

    def archive_run(self, run_id: str) -> dict:
        """Archives a run.

        Args:
            run_id: ID of the run.

        Returns:
            Empty dictionary.
        """
        return self._run_api.archive_run(run_id=run_id)

    def unarchive_run(self, run_id: str) -> dict:
        """Restores an archived run.

        Args:
            run_id: ID of the run.

        Returns:
            Empty dictionary.
        """
        return self._run_api.unarchive_run(run_id=run_id)

    def delete_run(self, run_id: str) -> dict:
        """Deletes a run.

        Args:
            run_id: ID of the run.

        Returns:
            Empty dictionary.
        """
        return self._run_api.delete_run(run_id=run_id)

    def terminate_run(self, run_id: str) -> dict:
        """Terminates a run.

        Args:
            run_id: ID of the run.

        Returns:
            Empty dictionary.
        """
        return self._run_api.terminate_run(run_id=run_id)

    def create_recurring_run(
        self,
        experiment_id: str,
        job_name: str,
        description: Optional[str] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        interval_second: Optional[int] = None,
        cron_expression: Optional[str] = None,
        max_concurrency: Optional[int] = 1,
        no_catchup: Optional[bool] = None,
        params: Optional[dict] = None,
        pipeline_package_path: Optional[str] = None,
        pipeline_id: Optional[str] = None,
        version_id: Optional[str] = None,
        enabled: bool = True,
        pipeline_root: Optional[str] = None,
        enable_caching: Optional[bool] = None,
        service_account: Optional[str] = None,
    ) -> kfp_server_api.V2beta1RecurringRun:
        """Creates a recurring run.

        Args:
            experiment_id: ID of the experiment.
            job_name: Name of the job.
            description: Description of the job.
            start_time: RFC3339 time string of the time when to start the
                job.
            end_time: RFC3339 time string of the time when to end the job.
            interval_second: Integer indicating the seconds between two
                recurring runs in for a periodic schedule.
            cron_expression: Cron expression representing a set of times,
                using 6 space-separated fields (e.g., ``'0 0 9 ? * 2-6'``). See `cron format
                <https://pkg.go.dev/github.com/robfig/cron#hdr-CRON_Expression_Format>`_.
            max_concurrency: Integer indicating how many jobs can be run in
                parallel.
            no_catchup: Whether the recurring run should catch up if behind
                schedule. For example, if the recurring run is paused for a
                while and re-enabled afterwards. If ``no_catchup=False``, the
                scheduler will catch up on (backfill) each missed interval.
                Otherwise, it only schedules the latest interval if more than
                one interval is ready to be scheduled. Usually, if your pipeline
                handles backfill internally, you should turn catchup off to
                avoid duplicate backfill.
            pipeline_package_path: Local path of the pipeline package (the
                filename should end with one of the following .tar.gz, .tgz,
                .zip, .json).
            params: Arguments to the pipeline function provided as a dict.
            pipeline_id: ID of a pipeline.
            version_id: ID of a pipeline version.
                If both ``pipeline_id`` and ``version_id`` are specified, ``version_id``
                will take precedence.
                If only ``pipeline_id`` is specified, the default version of this
                pipeline is used to create the run.
            enabled: Whether to enable or disable the recurring run.
            pipeline_root: Root path of the pipeline outputs.
            enable_caching: Whether or not to enable caching for the
                run. If not set, defaults to the compile time settings, which
                is ``True`` for all tasks by default, while users may specify
                different caching options for individual tasks. If set, the
                setting applies to all tasks in the pipeline (overrides the
                compile time settings).
            service_account: Specifies which Kubernetes service
                account this recurring run uses.
        Returns:
            ``V2beta1RecurringRun`` object.
        """

        job_config = self._create_job_config(
            params=params,
            pipeline_package_path=pipeline_package_path,
            pipeline_id=pipeline_id,
            version_id=version_id,
            enable_caching=enable_caching,
            pipeline_root=pipeline_root,
        )

        if all([interval_second, cron_expression
               ]) or not any([interval_second, cron_expression]):
            raise ValueError(
                'Either interval_second or cron_expression is required.')
        if interval_second is not None:
            trigger = kfp_server_api.V2beta1Trigger(
                periodic_schedule=kfp_server_api.V2beta1PeriodicSchedule(
                    start_time=start_time,
                    end_time=end_time,
                    interval_second=interval_second))
        if cron_expression is not None:
            trigger = kfp_server_api.V2beta1Trigger(
                cron_schedule=kfp_server_api.V2beta1CronSchedule(
                    start_time=start_time,
                    end_time=end_time,
                    cron=cron_expression))

        mode = kfp_server_api.RecurringRunMode.DISABLE
        if enabled:
            mode = kfp_server_api.RecurringRunMode.ENABLE

        job_body = kfp_server_api.V2beta1RecurringRun(
            experiment_id=experiment_id,
            mode=mode,
            pipeline_spec=job_config.pipeline_spec,
            pipeline_version_reference=job_config.pipeline_version_reference,
            runtime_config=job_config.runtime_config,
            display_name=job_name,
            description=description,
            no_catchup=no_catchup,
            trigger=trigger,
            max_concurrency=max_concurrency,
            service_account=service_account)
        return self._recurring_run_api.create_recurring_run(body=job_body)

    def _create_job_config(
        self,
        params: Optional[Dict[str, Any]],
        pipeline_package_path: Optional[str],
        pipeline_id: Optional[str],
        version_id: Optional[str],
        enable_caching: Optional[bool],
        pipeline_root: Optional[str],
    ) -> _JobConfig:
        """Creates a JobConfig with spec and resource_references.

        Args:
            pipeline_package_path: Local path of the pipeline package (the
                filename should end with one of the following .tar.gz, .tgz,
                .zip, .yaml, .yml).
            params: A dictionary with key as param name and value as param value.
            pipeline_id: ID of a pipeline.
            version_id: ID of a pipeline version.
                If both pipeline_id and version_id are specified, version_id
                will take precedence. If only pipeline_id is specified, the
                default version of this pipeline is used to create the run.
            enable_caching: Whether or not to enable caching for the
                run. If not set, defaults to the compile time settings, which
                is ``True`` for all tasks by default, while users may specify
                different caching options for individual tasks. If set, the
                setting applies to all tasks in the pipeline (overrides the
                compile time settings).
            pipeline_root: Root path of the pipeline outputs.

        Returns:
            A _JobConfig object with attributes .pipeline_spec,
                .pipeline_version_reference, and .runtime_config.
        """
        from_spec = pipeline_package_path is not None
        from_template = pipeline_id is not None or version_id is not None
        if from_spec == from_template:
            raise ValueError(
                'Must specify either `pipeline_pacakge_path` or both `pipeline_id` and `version_id`.'
            )
        if (pipeline_id is None) != (version_id is None):
            raise ValueError(
                'To run a pipeline from an existing template, both `pipeline_id` and `version_id` are required.'
            )

        if params is None:
            params = {}

        pipeline_spec = None
        if pipeline_package_path:
            pipeline_doc = _extract_pipeline_yaml(pipeline_package_path)

            # Caching option set at submission time overrides the compile time
            # settings.
            if enable_caching is not None:
                _override_caching_options(pipeline_doc.pipeline_spec,
                                          enable_caching)
            pipeline_spec = pipeline_doc.to_dict()

        pipeline_version_reference = None
        if pipeline_id is not None and version_id is not None:
            pipeline_version_reference = kfp_server_api.V2beta1PipelineVersionReference(
                pipeline_id=pipeline_id, pipeline_version_id=version_id)

        runtime_config = kfp_server_api.V2beta1RuntimeConfig(
            pipeline_root=pipeline_root,
            parameters=params,
        )
        return _JobConfig(
            pipeline_spec=pipeline_spec,
            pipeline_version_reference=pipeline_version_reference,
            runtime_config=runtime_config,
        )

    def create_run_from_pipeline_func(
        self,
        pipeline_func: base_component.BaseComponent,
        arguments: Optional[Dict[str, Any]] = None,
        run_name: Optional[str] = None,
        experiment_name: Optional[str] = None,
        namespace: Optional[str] = None,
        pipeline_root: Optional[str] = None,
        enable_caching: Optional[bool] = None,
        service_account: Optional[str] = None,
        experiment_id: Optional[str] = None,
    ) -> RunPipelineResult:
        """Runs pipeline on KFP-enabled Kubernetes cluster.

        This command compiles the pipeline function, creates or gets an
        experiment, then submits the pipeline for execution.

        Args:
            pipeline_func: Pipeline function constructed with ``@kfp.dsl.pipeline`` decorator.
            arguments: Arguments to the pipeline function provided as a dict.
            run_name: Name of the run to be shown in the UI.
            experiment_name: Name of the experiment to add the run to. You cannot specify both experiment_name and experiment_id.
            namespace: Kubernetes namespace to use. Used for multi-user deployments. For single-user deployments, this should be left as ``None``.
            pipeline_root: Root path of the pipeline outputs.
            enable_caching: Whether or not to enable caching for the
                run. If not set, defaults to the compile time settings, which
                is ``True`` for all tasks by default, while users may specify
                different caching options for individual tasks. If set, the
                setting applies to all tasks in the pipeline (overrides the
                compile time settings).
            service_account: Specifies which Kubernetes service
                account to use for this run.
            experiment_id: ID of the experiment to add the run to. You cannot specify both experiment_id and experiment_name.

        Returns:
            ``RunPipelineResult`` object containing information about the pipeline run.
        """
        #TODO: Check arguments against the pipeline function
        pipeline_name = pipeline_func.name
        run_name = run_name or pipeline_name + ' ' + datetime.datetime.now(
        ).strftime('%Y-%m-%d %H-%M-%S')

        with tempfile.TemporaryDirectory() as tmpdir:
            pipeline_package_path = os.path.join(tmpdir, 'pipeline.yaml')
            compiler.Compiler().compile(
                pipeline_func=pipeline_func,
                package_path=pipeline_package_path,
            )

            return self.create_run_from_pipeline_package(
                pipeline_file=pipeline_package_path,
                arguments=arguments,
                run_name=run_name,
                experiment_id=experiment_id,
                experiment_name=experiment_name,
                namespace=namespace,
                pipeline_root=pipeline_root,
                enable_caching=enable_caching,
                service_account=service_account,
            )

    def create_run_from_pipeline_package(
        self,
        pipeline_file: str,
        arguments: Optional[Dict[str, Any]] = None,
        run_name: Optional[str] = None,
        experiment_name: Optional[str] = None,
        namespace: Optional[str] = None,
        pipeline_root: Optional[str] = None,
        enable_caching: Optional[bool] = None,
        service_account: Optional[str] = None,
        experiment_id: Optional[str] = None,
    ) -> RunPipelineResult:
        """Runs pipeline on KFP-enabled Kubernetes cluster.

        This command takes a local pipeline package, creates or gets an
        experiment, then submits the pipeline for execution.

        Args:
            pipeline_file: A compiled pipeline package file.
            arguments: Arguments to the pipeline function provided as a dict.
            run_name:  Name of the run to be shown in the UI.
            experiment_name: Name of the experiment to add the run to. You cannot specify both experiment_name and experiment_id.
            namespace: Kubernetes namespace to use. Used for multi-user deployments. For single-user deployments, this should be left as ``None``.
            pipeline_root: Root path of the pipeline outputs.
            enable_caching: Whether or not to enable caching for the
                run. If not set, defaults to the compile time settings, which
                is ``True`` for all tasks by default, while users may specify
                different caching options for individual tasks. If set, the
                setting applies to all tasks in the pipeline (overrides the
                compile time settings).
            service_account: Specifies which Kubernetes service
                account to use for this run.
            experiment_id: ID of the experiment to add the run to. You cannot specify both experiment_id and experiment_name.

        Returns:
            ``RunPipelineResult`` object containing information about the pipeline run.
        """

        #TODO: Check arguments against the pipeline function
        pipeline_name = os.path.basename(pipeline_file)

        if (experiment_name is not None) and (experiment_id is not None):
            raise ValueError(
                'You cannot specify both experiment_name and experiment_id.')

        if not experiment_id:
            experiment_name = experiment_name or os.environ.get(
                KF_PIPELINES_DEFAULT_EXPERIMENT_NAME, None)
            overridden_experiment_name = os.environ.get(
                KF_PIPELINES_OVERRIDE_EXPERIMENT_NAME, experiment_name)
            if overridden_experiment_name != experiment_name:
                warnings.warn(
                    f'Changing experiment name from "{experiment_name}" to "{overridden_experiment_name}".'
                )
            experiment_name = overridden_experiment_name or 'Default'
            experiment = self.create_experiment(
                name=experiment_name, namespace=namespace)
            experiment_id = experiment.experiment_id

        run_name = run_name or (
            pipeline_name + ' ' +
            datetime.datetime.now().strftime('%Y-%m-%d %H-%M-%S'))

        run_info = self.run_pipeline(
            experiment_id=experiment_id,
            job_name=run_name,
            pipeline_package_path=pipeline_file,
            params=arguments,
            pipeline_root=pipeline_root,
            enable_caching=enable_caching,
            service_account=service_account,
        )
        return RunPipelineResult(self, run_info)

    def delete_job(self, job_id: str) -> dict:
        """Deletes a job (recurring run).

        Args:
            job_id: ID of the job.

        Returns:
            Empty dictionary.
        """
        warnings.warn(
            '`delete_job` is deprecated. Please use `delete_recurring_run` instead.'
            f'\nReroute to calling `delete_recurring_run(recurring_run_id="{job_id}")`',
            category=DeprecationWarning,
            stacklevel=2)
        return self.delete_recurring_run(recurring_run_id=job_id)

    def delete_recurring_run(self, recurring_run_id: str) -> dict:
        """Deletes a recurring run.

        Args:
            recurring_run_id: ID of the recurring_run.

        Returns:
            Empty dictionary.
        """
        return self._recurring_run_api.delete_recurring_run(
            recurring_run_id=recurring_run_id)

    def disable_job(self, job_id: str) -> dict:
        """Disables a job (recurring run).

        Args:
            job_id: ID of the job.

        Returns:
            Empty dictionary.
        """
        warnings.warn(
            '`disable_job` is deprecated. Please use `disable_recurring_run` instead.'
            f'\nReroute to calling `disable_recurring_run(recurring_run_id="{job_id}")`',
            category=DeprecationWarning,
            stacklevel=2)
        return self.disable_recurring_run(recurring_run_id=job_id)

    def disable_recurring_run(self, recurring_run_id: str) -> dict:
        """Disables a recurring run.

        Args:
            recurring_run_id: ID of the recurring_run.

        Returns:
            Empty dictionary.
        """
        return self._recurring_run_api.disable_recurring_run(
            recurring_run_id=recurring_run_id)

    def enable_job(self, job_id: str) -> dict:
        """Enables a job (recurring run).

        Args:
            job_id: ID of the job.

        Returns:
            Empty dictionary.
        """
        warnings.warn(
            '`enable_job` is deprecated. Please use `enable_recurring_run` instead.'
            f'\nReroute to calling `enable_recurring_run(recurring_run_id="{job_id}")`',
            category=DeprecationWarning,
            stacklevel=2)
        return self.enable_recurring_run(recurring_run_id=job_id)

    def enable_recurring_run(self, recurring_run_id: str) -> dict:
        """Enables a recurring run.

        Args:
            recurring_run_id: ID of the recurring_run.

        Returns:
            Empty dictionary.
        """
        return self._recurring_run_api.enable_recurring_run(
            recurring_run_id=recurring_run_id)

    def list_runs(
        self,
        page_token: str = '',
        page_size: int = 10,
        sort_by: str = '',
        experiment_id: Optional[str] = None,
        namespace: Optional[str] = None,
        filter: Optional[str] = None,
    ) -> kfp_server_api.V2beta1ListRunsResponse:
        """List runs.

        Args:
            page_token: Page token for obtaining page from paginated response.
            page_size: Size of the page.
            sort_by: Sort string of format ``'[field_name]', '[field_name] desc'``. For example, ``'display_name desc'``.
            experiment_id: Experiment ID to filter upon
            namespace: Kubernetes namespace to use. Used for multi-user deployments. For single-user deployments, this should be left as ``None``.
            filter: A url-encoded, JSON-serialized Filter protocol buffer
                (see `filter.proto message <https://github.com/kubeflow/pipelines/blob/cb7d9a87c999eb1d2280959e5afbeee9e270ef3d/backend/api/v2beta1/filter.proto>`_). For a list of all filter operations ``'opertion'``, see `here <https://github.com/kubeflow/pipelines/blob/777c98153daf3dfae82730e14ff37bdddc334c4d/sdk/python/kfp/client/client.py#L37-L45>`_. Example:

                  ::

                    json.dumps({
                        "predicates": [{
                            "operation": "EQUALS",
                            "key": "display_name",
                            "stringValue": "my-name",
                        }]
                    })

          Returns:
            ``V2beta1ListRunsResponse`` object.
        """
        namespace = namespace or self.get_user_namespace()
        if experiment_id is not None:
            return self._run_api.list_runs(
                page_token=page_token,
                page_size=page_size,
                sort_by=sort_by,
                experiment_id=experiment_id,
                filter=filter)

        elif namespace is not None:
            return self._run_api.list_runs(
                page_token=page_token,
                page_size=page_size,
                sort_by=sort_by,
                namespace=namespace,
                filter=filter)

        else:
            return self._run_api.list_runs(
                page_token=page_token,
                page_size=page_size,
                sort_by=sort_by,
                filter=filter)

    def list_recurring_runs(
        self,
        page_token: str = '',
        page_size: int = 10,
        sort_by: str = '',
        experiment_id: Optional[str] = None,
        namespace: Optional[str] = None,
        filter: Optional[str] = None,
    ) -> kfp_server_api.V2beta1ListRecurringRunsResponse:
        """Lists recurring runs.

        Args:
            page_token: Page token for obtaining page from paginated response.
            page_size: Size of the page.
            sort_by: Sort string of format ``'[field_name]', '[field_name] desc'``. For example, ``'display_name desc'``.
            experiment_id: Experiment ID to filter upon.
            namespace: Kubernetes namespace to use. Used for multi-user deployments. For single-user deployments, this should be left as ``None``.
            filter: A url-encoded, JSON-serialized Filter protocol buffer
                (see `filter.proto message <https://github.com/kubeflow/pipelines/blob/cb7d9a87c999eb1d2280959e5afbeee9e270ef3d/backend/api/v2beta1/filter.proto>`_). Example:

                  ::

                    json.dumps({
                        "predicates": [{
                            "operation": "EQUALS",
                            "key": "display_name",
                            "stringValue": "my-name",
                        }]
                    })

        Returns:
            ``V2beta1ListRecurringRunsResponse`` object.
        """
        if experiment_id is not None:
            return self._recurring_run_api.list_recurring_runs(
                page_token=page_token,
                page_size=page_size,
                sort_by=sort_by,
                experiment_id=experiment_id,
                filter=filter)

        elif namespace is not None:
            return self._recurring_run_api.list_recurring_runs(
                page_token=page_token,
                page_size=page_size,
                sort_by=sort_by,
                namespace=namespace,
                filter=filter)

        else:
            return self._recurring_run_api.list_recurring_runs(
                page_token=page_token,
                page_size=page_size,
                sort_by=sort_by,
                filter=filter)

    def get_recurring_run(
        self,
        recurring_run_id: str,
        job_id: Optional[str] = None,
    ) -> kfp_server_api.V2beta1RecurringRun:
        """Gets recurring run details.

        Args:
            recurring_run_id: ID of the recurring run.
            job_id: Deprecated. Use `recurring_run_id` instead.

        Returns:
            ``V2beta1RecurringRun`` object.
        """
        if job_id is not None:
            warnings.warn(
                '`job_id` is deprecated. Please use `recurring_run_id` instead.',
                category=DeprecationWarning,
                stacklevel=2)
            recurring_run_id = recurring_run_id or job_id

        return self._recurring_run_api.get_recurring_run(
            recurring_run_id=recurring_run_id)

    def get_run(self, run_id: str) -> kfp_server_api.V2beta1Run:
        """Gets run details.

        Args:
            run_id: ID of the run.

        Returns:
            ``V2beta1Run`` object.
        """
        return self._run_api.get_run(run_id=run_id)

    def wait_for_run_completion(
        self,
        run_id: str,
        timeout: int,
        sleep_duration: int = 5,
    ) -> kfp_server_api.V2beta1Run:
        """Waits for a run to complete.

        Args:
            run_id: ID of the run.
            timeout: Timeout after which the client should stop waiting for run completion (seconds).
            sleep_duration: Time in seconds between retries.

        Returns:
            ``V2beta1Run`` object.
        """
        state = 'Running:'
        start_time = datetime.datetime.now()
        if isinstance(timeout, datetime.timedelta):
            timeout = timeout.total_seconds()
        is_valid_token = False
        finish_states = ['succeeded', 'failed', 'skipped', 'error']
        while True:
            try:
                get_run_response = self._run_api.get_run(run_id=run_id)
                is_valid_token = True
            except kfp_server_api.ApiException as api_ex:
                # if the token is valid but receiving 401 Unauthorized error
                # then refresh the token
                if is_valid_token and api_ex.status == 401:
                    logging.info('Access token has expired !!! Refreshing ...')
                    self._refresh_api_client_token()
                    continue
                else:
                    raise api_ex
            state = get_run_response.state
            elapsed_time = (datetime.datetime.now() -
                            start_time).total_seconds()
            logging.info('Waiting for the job to complete...')
            if elapsed_time > timeout:
                raise TimeoutError('Run timeout')
            if state is not None and state.lower() in finish_states:
                return get_run_response
            time.sleep(sleep_duration)

    def upload_pipeline(
        self,
        pipeline_package_path: str,
        pipeline_name: Optional[str] = None,
        description: Optional[str] = None,
        namespace: Optional[str] = None,
    ) -> kfp_server_api.V2beta1Pipeline:
        """Uploads a pipeline.

        Args:
            pipeline_package_path: Local path to the pipeline package.
            pipeline_name: Name of the pipeline to be shown in the UI.
            description: Description of the pipeline to be shown in the UI.
            namespace: Optional. Kubernetes namespace where the pipeline should
                be uploaded. For single user deployment, leave it as None; For
                multi user, input a namespace where the user is authorized.

        Returns:
            ``V2beta1Pipeline`` object.
        """
        if pipeline_name is None:
            pipeline_doc = _extract_pipeline_yaml(pipeline_package_path)
            pipeline_name = pipeline_doc.pipeline_spec.pipeline_info.name
        validate_pipeline_resource_name(pipeline_name)
        response = self._upload_api.upload_pipeline(
            pipeline_package_path,
            name=pipeline_name,
            description=description,
            namespace=namespace)
        link = f'{self._get_url_prefix()}/#/pipelines/details/{response.pipeline_id}'
        if self._is_ipython():
            import IPython
            html = f'<a href="{link}" target="_blank" >Pipeline details</a>.'
            IPython.display.display(IPython.display.HTML(html))
        else:
            print(f'Pipeline details: {link}')

        return response

    def upload_pipeline_version(
        self,
        pipeline_package_path: str,
        pipeline_version_name: str,
        pipeline_id: Optional[str] = None,
        pipeline_name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> kfp_server_api.V2beta1PipelineVersion:
        """Uploads a new version of the pipeline.

        Args:
            pipeline_package_path: Local path to the pipeline package.
            pipeline_version_name:  Name of the pipeline version to be shown in
                the UI.
            pipeline_id: ID of the pipeline.
            pipeline_name: Name of the pipeline.
            description: Description of the pipeline version to show in the UI.

        Returns:
            ``V2beta1PipelineVersion`` object.
        """

        if all([pipeline_id, pipeline_name
               ]) or not any([pipeline_id, pipeline_name]):
            raise ValueError('Either pipeline_id or pipeline_name is required.')

        if pipeline_name:
            pipeline_id = self.get_pipeline_id(pipeline_name)
        kwargs = dict(
            name=pipeline_version_name,
            pipelineid=pipeline_id,
        )

        if description:
            kwargs['description'] = description

        response = self._upload_api.upload_pipeline_version(
            pipeline_package_path, **kwargs)

        link = f'{self._get_url_prefix()}/#/pipelines/details/{response.pipeline_id}/version/{response.pipeline_version_id}'
        if self._is_ipython():
            import IPython
            html = f'<a href="{link}" target="_blank" >Pipeline details</a>.'
            IPython.display.display(IPython.display.HTML(html))
        else:
            print(f'Pipeline details: {link}')

        return response

    def get_pipeline(self, pipeline_id: str) -> kfp_server_api.V2beta1Pipeline:
        """Gets pipeline details.

        Args:
            pipeline_id: ID of the pipeline.

        Returns:
            ``V2beta1Pipeline`` object.
        """
        return self._pipelines_api.get_pipeline(pipeline_id=pipeline_id)

    def delete_pipeline(self, pipeline_id: str) -> dict:
        """Deletes a pipeline.

        Args:
            pipeline_id: ID of the pipeline.

        Returns:
            Empty dictionary.
        """
        return self._pipelines_api.delete_pipeline(pipeline_id=pipeline_id)

    def list_pipeline_versions(
        self,
        pipeline_id: str,
        page_token: str = '',
        page_size: int = 10,
        sort_by: str = '',
        filter: Optional[str] = None,
    ) -> kfp_server_api.V2beta1ListPipelineVersionsResponse:
        """Lists pipeline versions.

        Args:
            pipeline_id: ID of the pipeline for which to list versions.
            page_token: Page token for obtaining page from paginated response.
            page_size: Size of the page.
            sort_by: Sort string of format ``'[field_name]', '[field_name] desc'``. For example, ``'display_name desc'``.
            filter: A url-encoded, JSON-serialized Filter protocol buffer
                (see `filter.proto message <https://github.com/kubeflow/pipelines/blob/cb7d9a87c999eb1d2280959e5afbeee9e270ef3d/backend/api/v2beta1/filter.proto>`_). Example:

                  ::

                    json.dumps({
                        "predicates": [{
                            "operation": "EQUALS",
                            "key": "display_name",
                            "stringValue": "my-name",
                        }]
                    })

        Returns:
            ``V2beta1ListPipelineVersionsResponse`` object.
        """

        return self._pipelines_api.list_pipeline_versions(
            page_token=page_token,
            page_size=page_size,
            sort_by=sort_by,
            pipeline_id=pipeline_id,
            filter=filter)

    def get_pipeline_version(
        self,
        pipeline_id: str,
        pipeline_version_id: str,
    ) -> kfp_server_api.V2beta1PipelineVersion:
        """Gets a pipeline version.

        Args:
            pipeline_id: ID of the pipeline.
            pipeline_version_id: ID of the pipeline version.

        Returns:
            ``V2beta1PipelineVersion`` object.
        """
        return self._pipelines_api.get_pipeline_version(
            pipeline_id=pipeline_id,
            pipeline_version_id=pipeline_version_id,
        )

    def delete_pipeline_version(
        self,
        pipeline_id: str,
        pipeline_version_id: str,
    ) -> dict:
        """Deletes a pipeline version.p.

        Args:
            pipeline_id: ID of the pipeline.
            pipeline_version_id: ID of the pipeline version.

        Returns:
            Empty dictionary.
        """
        return self._pipelines_api.delete_pipeline_version(
            pipeline_id=pipeline_id,
            pipeline_version_id=pipeline_version_id,
        )


def _add_generated_apis(
    target_struct: Any,
    api_module: ModuleType,
    api_client: kfp_server_api.ApiClient,
) -> None:
    """Initializes a hierarchical API object based on the generated API module.

    PipelineServiceApi.create_pipeline becomes
    target_struct.pipelines.create_pipeline
    """
    Struct = type('Struct', (), {})

    def camel_case_to_snake_case(name: str) -> str:
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()

    for api_name in dir(api_module):
        if not api_name.endswith('ServiceApi'):
            continue

        short_api_name = camel_case_to_snake_case(
            api_name[0:-len('ServiceApi')]) + 's'
        api_struct = Struct()
        setattr(target_struct, short_api_name, api_struct)
        service_api = getattr(api_module.api, api_name)
        initialized_service_api = service_api(api_client)
        for member_name in dir(initialized_service_api):
            if member_name.startswith('_') or member_name.endswith(
                    '_with_http_info'):
                continue

            bound_member = getattr(initialized_service_api, member_name)
            setattr(api_struct, member_name, bound_member)
    models_struct = Struct()
    for member_name in dir(api_module.models):
        if not member_name[0].islower():
            setattr(models_struct, member_name,
                    getattr(api_module.models, member_name))
    target_struct.api_models = models_struct


def validate_pipeline_resource_name(name: str) -> None:
    REGEX = r'[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*'

    if re.fullmatch(REGEX, name) is None:
        raise ValueError(
            f'Invalid pipeline name: "{name}". Pipeline name must conform to the regex: "{REGEX}".'
        )


def _extract_pipeline_yaml(package_file: str) -> _PipelineDoc:

    def _choose_pipeline_file(file_list: List[str]) -> str:
        pipeline_files = [file for file in file_list if file.endswith('.yaml')]
        if not pipeline_files:
            raise ValueError(
                'Invalid package. Missing pipeline yaml file in the package.')

        if 'pipeline.yaml' in pipeline_files:
            return 'pipeline.yaml'
        elif len(pipeline_files) == 1:
            return pipeline_files[0]
        else:
            raise ValueError(
                'Invalid package. There is no pipeline.json file or there '
                'are multiple yaml files.')

    def _safe_load_yaml(stream: TextIO) -> _PipelineDoc:
        docs = yaml.safe_load_all(stream)
        pipeline_spec_dict = None
        platform_spec_dict = {}
        for doc in docs:
            if pipeline_spec_dict is None:
                pipeline_spec_dict = doc
            else:
                platform_spec_dict.update(doc)

        return _PipelineDoc(
            pipeline_spec=json_format.ParseDict(
                pipeline_spec_dict, pipeline_spec_pb2.PipelineSpec()),
            platform_spec=json_format.ParseDict(
                platform_spec_dict, pipeline_spec_pb2.PlatformSpec()))

    if package_file.endswith('.tar.gz') or package_file.endswith('.tgz'):
        with tarfile.open(package_file, 'r:gz') as tar:
            file_names = [member.name for member in tar if member.isfile()]
            pipeline_file = _choose_pipeline_file(file_names)
            with tar.extractfile(
                    tar.getmember(pipeline_file)) as f:  # type: ignore
                return _safe_load_yaml(f)
    elif package_file.endswith('.zip'):
        with zipfile.ZipFile(package_file, 'r') as zip:
            pipeline_file = _choose_pipeline_file(zip.namelist())
            with zip.open(pipeline_file) as f:
                return _safe_load_yaml(f)
    elif package_file.endswith('.yaml') or package_file.endswith('.yml'):
        with open(package_file, 'r') as f:
            return _safe_load_yaml(f)
    else:
        raise ValueError(
            f'The package_file {package_file} should end with one of the '
            'following formats: [.tar.gz, .tgz, .zip, .yaml, .yml].')


def _override_caching_options(
    pipeline_spec: pipeline_spec_pb2.PipelineSpec,
    enable_caching: bool,
) -> None:
    """Overrides caching options.

    Args:
        pipeline_spec: The PipelineSpec object to update in-place.
        enable_caching: Overrides options, one of True, False.
    """
    for _, task_spec in pipeline_spec.root.dag.tasks.items():
        task_spec.caching_options.enable_cache = enable_caching



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for creating IR ComponentSpec instance."""

from typing import List, Optional, Tuple, Union

from kfp.deprecated.components import _structures as structures
from kfp.deprecated.dsl import _for_loop, _pipeline_param, dsl_utils
from kfp.pipeline_spec import pipeline_spec_pb2
from kfp.deprecated.dsl import type_utils


def additional_input_name_for_pipelineparam(
        param_or_name: Union[_pipeline_param.PipelineParam, str]) -> str:
    """Gets the name for an additional (compiler-injected) input."""

    # Adding a prefix to avoid (reduce chance of) name collision between the
    # original component inputs and the injected input.
    return 'pipelineparam--' + (
        param_or_name.full_name if isinstance(
            param_or_name, _pipeline_param.PipelineParam) else param_or_name)


def _exclude_loop_arguments_variables(
    param_or_name: Union[_pipeline_param.PipelineParam, str]
) -> Tuple[str, Optional[str]]:
    """Gets the pipeline param name excluding loop argument variables.

    Args:
      param: The pipeline param object which may or may not be a loop argument.

    Returns:
      A tuple of the name of the pipeline param without loop arguments subvar name
      and the subvar name is found.
    """
    if isinstance(param_or_name, _pipeline_param.PipelineParam):
        param_name = param_or_name.full_name
    else:
        param_name = param_or_name

    subvar_name = None
    # Special handling for loop arguments.
    # In case of looping over a list of maps, each subvar (a key in the map)
    # referencing yields a pipeline param. For example:
    # - some-param-loop-item: referencing the whole map
    # - some-param-loop-item-subvar-key1: referencing key1 in the map.
    # Because of the way IR is designed to support looping over a subvar (using
    # `parameter_expression_selector`), we don't create inputs for subvariables.
    # So if we see a pipeline param named 'some-param-loop-item-subvar-key1',
    # we build the component_spec/task_spec inputs using 'some-param-loop-item'
    # (without the subvar suffix).
    if _for_loop.LoopArguments.name_is_loop_argument(param_name):
        # Subvar pipeline params may not have types, defaults to string type.
        if isinstance(param_or_name, _pipeline_param.PipelineParam):
            param_or_name.param_type = param_or_name.param_type or 'String'
        loop_args_name_and_var_name = (
            _for_loop.LoopArgumentVariable
            .parse_loop_args_name_and_this_var_name(param_name))
        if loop_args_name_and_var_name:
            param_name = loop_args_name_and_var_name[0]
            subvar_name = loop_args_name_and_var_name[1]
    return (param_name, subvar_name)


def build_component_spec_from_structure(
    component_spec: structures.ComponentSpec,
    executor_label: str,
    actual_inputs: List[str],
) -> pipeline_spec_pb2.ComponentSpec:
    """Builds an IR ComponentSpec instance from structures.ComponentSpec.

    Args:
      component_spec: The structure component spec.
      executor_label: The executor label.
      actual_inputs: The actual arugments passed to the task. This is used as a
        short term workaround to support optional inputs in component spec IR.

    Returns:
      An instance of IR ComponentSpec.
    """
    result = pipeline_spec_pb2.ComponentSpec()
    result.executor_label = executor_label

    for input_spec in component_spec.inputs or []:
        # skip inputs not present
        if input_spec.name not in actual_inputs:
            continue
        if type_utils.is_parameter_type(input_spec.type):
            result.input_definitions.parameters[
                input_spec.name].parameter_type = type_utils.get_parameter_type(
                    input_spec.type)
        else:
            result.input_definitions.artifacts[
                input_spec.name].artifact_type.CopyFrom(
                    type_utils.get_artifact_type_schema(input_spec.type))

    for output_spec in component_spec.outputs or []:
        if type_utils.is_parameter_type(output_spec.type):
            result.output_definitions.parameters[
                output_spec
                .name].parameter_type = type_utils.get_parameter_type(
                    output_spec.type)
        else:
            result.output_definitions.artifacts[
                output_spec.name].artifact_type.CopyFrom(
                    type_utils.get_artifact_type_schema(output_spec.type))

    return result


def build_component_inputs_spec(
    component_spec: pipeline_spec_pb2.ComponentSpec,
    pipeline_params: List[_pipeline_param.PipelineParam],
    is_root_component: bool,
) -> None:
    """Builds component inputs spec from pipeline params.

    Args:
      component_spec: The component spec to fill in its inputs spec.
      pipeline_params: The list of pipeline params.
      is_root_component: Whether the component is the root.
    """
    for param in pipeline_params:
        param_name = param.full_name
        if _for_loop.LoopArguments.name_is_loop_argument(param_name):
            param.param_type = param.param_type or 'String'

        input_name = (
            param_name if is_root_component else
            additional_input_name_for_pipelineparam(param_name))

        if type_utils.is_parameter_type(param.param_type):
            component_spec.input_definitions.parameters[
                input_name].parameter_type = type_utils.get_parameter_type(
                    param.param_type)
        elif input_name not in getattr(component_spec.input_definitions,
                                       'parameters', []):
            component_spec.input_definitions.artifacts[
                input_name].artifact_type.CopyFrom(
                    type_utils.get_artifact_type_schema(param.param_type))


def build_component_outputs_spec(
    component_spec: pipeline_spec_pb2.ComponentSpec,
    pipeline_params: List[_pipeline_param.PipelineParam],
) -> None:
    """Builds component outputs spec from pipeline params.

    Args:
      component_spec: The component spec to fill in its outputs spec.
      pipeline_params: The list of pipeline params.
    """
    for param in pipeline_params or []:
        output_name = param.full_name
        if type_utils.is_parameter_type(param.param_type):
            component_spec.output_definitions.parameters[
                output_name].parameter_type = type_utils.get_parameter_type(
                    param.param_type)
        elif output_name not in getattr(component_spec.output_definitions,
                                        'parameters', []):
            component_spec.output_definitions.artifacts[
                output_name].artifact_type.CopyFrom(
                    type_utils.get_artifact_type_schema(param.param_type))


def build_task_inputs_spec(
    task_spec: pipeline_spec_pb2.PipelineTaskSpec,
    pipeline_params: List[_pipeline_param.PipelineParam],
    tasks_in_current_dag: List[str],
    is_parent_component_root: bool,
) -> None:
    """Builds task inputs spec from pipeline params.

    Args:
      task_spec: The task spec to fill in its inputs spec.
      pipeline_params: The list of pipeline params.
      tasks_in_current_dag: The list of tasks names for tasks in the same dag.
      is_parent_component_root: Whether the task is in the root component.
    """
    for param in pipeline_params or []:

        param_full_name, subvar_name = _exclude_loop_arguments_variables(param)
        input_name = additional_input_name_for_pipelineparam(param.full_name)

        param_name = param.name
        if subvar_name:
            task_spec.inputs.parameters[
                input_name].parameter_expression_selector = (
                    'parseJson(string_value)["{}"]'.format(subvar_name))
            param_name = _for_loop.LoopArguments.remove_loop_item_base_name(
                _exclude_loop_arguments_variables(param_name)[0])

        if type_utils.is_parameter_type(param.param_type):
            if param.op_name and dsl_utils.sanitize_task_name(
                    param.op_name) in tasks_in_current_dag:
                task_spec.inputs.parameters[
                    input_name].task_output_parameter.producer_task = (
                        dsl_utils.sanitize_task_name(param.op_name))
                task_spec.inputs.parameters[
                    input_name].task_output_parameter.output_parameter_key = (
                        param_name)
            else:
                task_spec.inputs.parameters[
                    input_name].component_input_parameter = (
                        param_full_name if is_parent_component_root else
                        additional_input_name_for_pipelineparam(param_full_name)
                    )
        else:
            if param.op_name and dsl_utils.sanitize_task_name(
                    param.op_name) in tasks_in_current_dag:
                task_spec.inputs.artifacts[
                    input_name].task_output_artifact.producer_task = (
                        dsl_utils.sanitize_task_name(param.op_name))
                task_spec.inputs.artifacts[
                    input_name].task_output_artifact.output_artifact_key = (
                        param_name)
            else:
                task_spec.inputs.artifacts[
                    input_name].component_input_artifact = (
                        param_full_name
                        if is_parent_component_root else input_name)


def update_task_inputs_spec(
    task_spec: pipeline_spec_pb2.PipelineTaskSpec,
    parent_component_inputs: pipeline_spec_pb2.ComponentInputsSpec,
    pipeline_params: List[_pipeline_param.PipelineParam],
    tasks_in_current_dag: List[str],
    input_parameters_in_current_dag: List[str],
    input_artifacts_in_current_dag: List[str],
) -> None:
    """Updates task inputs spec.

    A task input may reference an output outside its immediate DAG.
    For instance::

      random_num = random_num_op(...)
      with dsl.Condition(random_num.output > 5):
        print_op('%s > 5' % random_num.output)

    In this example, `dsl.Condition` forms a sub-DAG with one task from `print_op`
    inside the sub-DAG. The task of `print_op` references output from `random_num`
    task, which is outside the sub-DAG. When compiling to IR, such cross DAG
    reference is disallowed. So we need to "punch a hole" in the sub-DAG to make
    the input available in the sub-DAG component inputs if it's not already there,
    Next, we can call this method to fix the tasks inside the sub-DAG to make them
    reference the component inputs instead of directly referencing the original
    producer task.

    Args:
      task_spec: The task spec to fill in its inputs spec.
      parent_component_inputs: The input spec of the task's parent component.
      pipeline_params: The list of pipeline params.
      tasks_in_current_dag: The list of tasks names for tasks in the same dag.
      input_parameters_in_current_dag: The list of input parameters in the DAG
        component.
      input_artifacts_in_current_dag: The list of input artifacts in the DAG
        component.
    """
    if not hasattr(task_spec, 'inputs'):
        return

    for input_name in getattr(task_spec.inputs, 'parameters', []):

        if task_spec.inputs.parameters[input_name].WhichOneof(
                'kind') == 'task_output_parameter' and (
                    task_spec.inputs.parameters[input_name]
                    .task_output_parameter.producer_task
                    not in tasks_in_current_dag):

            param = _pipeline_param.PipelineParam(
                name=task_spec.inputs.parameters[input_name]
                .task_output_parameter.output_parameter_key,
                op_name=task_spec.inputs.parameters[input_name]
                .task_output_parameter.producer_task)

            component_input_parameter = (
                additional_input_name_for_pipelineparam(param.full_name))

            if component_input_parameter in parent_component_inputs.parameters:
                task_spec.inputs.parameters[
                    input_name].component_input_parameter = component_input_parameter
                continue

            # The input not found in parent's component input definitions
            # This could happen because of loop arguments variables
            param_name, subvar_name = _exclude_loop_arguments_variables(param)
            if subvar_name:
                task_spec.inputs.parameters[
                    input_name].parameter_expression_selector = (
                        'parseJson(string_value)["{}"]'.format(subvar_name))

            component_input_parameter = (
                additional_input_name_for_pipelineparam(param_name))

            assert component_input_parameter in parent_component_inputs.parameters, \
                'component_input_parameter: {} not found. All inputs: {}'.format(
                    component_input_parameter, parent_component_inputs)

            task_spec.inputs.parameters[
                input_name].component_input_parameter = component_input_parameter

        elif task_spec.inputs.parameters[input_name].WhichOneof(
                'kind') == 'component_input_parameter':

            component_input_parameter = (
                task_spec.inputs.parameters[input_name]
                .component_input_parameter)

            if component_input_parameter in parent_component_inputs.parameters:
                continue

            if additional_input_name_for_pipelineparam(
                    component_input_parameter
            ) in parent_component_inputs.parameters:
                task_spec.inputs.parameters[
                    input_name].component_input_parameter = (
                        additional_input_name_for_pipelineparam(
                            component_input_parameter))
                continue

            # The input not found in parent's component input definitions
            # This could happen because of loop arguments variables
            component_input_parameter, subvar_name = _exclude_loop_arguments_variables(
                component_input_parameter)

            if subvar_name:
                task_spec.inputs.parameters[
                    input_name].parameter_expression_selector = (
                        'parseJson(string_value)["{}"]'.format(subvar_name))

            if component_input_parameter not in input_parameters_in_current_dag:
                component_input_parameter = (
                    additional_input_name_for_pipelineparam(
                        component_input_parameter))

            if component_input_parameter not in parent_component_inputs.parameters:
                component_input_parameter = (
                    additional_input_name_for_pipelineparam(
                        component_input_parameter))
            assert component_input_parameter in parent_component_inputs.parameters, \
              'component_input_parameter: {} not found. All inputs: {}'.format(
                  component_input_parameter, parent_component_inputs)

            task_spec.inputs.parameters[
                input_name].component_input_parameter = component_input_parameter

    for input_name in getattr(task_spec.inputs, 'artifacts', []):

        if task_spec.inputs.artifacts[input_name].WhichOneof(
                'kind') == 'task_output_artifact' and (
                    task_spec.inputs.artifacts[input_name].task_output_artifact
                    .producer_task not in tasks_in_current_dag):

            param = _pipeline_param.PipelineParam(
                name=task_spec.inputs.artifacts[input_name].task_output_artifact
                .output_artifact_key,
                op_name=task_spec.inputs.artifacts[input_name]
                .task_output_artifact.producer_task)
            component_input_artifact = (
                additional_input_name_for_pipelineparam(param))
            assert component_input_artifact in parent_component_inputs.artifacts, \
              'component_input_artifact: {} not found. All inputs: {}'.format(
                  component_input_artifact, parent_component_inputs)

            task_spec.inputs.artifacts[
                input_name].component_input_artifact = component_input_artifact

        elif task_spec.inputs.artifacts[input_name].WhichOneof(
                'kind') == 'component_input_artifact':

            component_input_artifact = (
                task_spec.inputs.artifacts[input_name].component_input_artifact)

            if component_input_artifact not in input_artifacts_in_current_dag:
                component_input_artifact = (
                    additional_input_name_for_pipelineparam(
                        task_spec.inputs.artifacts[input_name]
                        .component_input_artifact))
                assert component_input_artifact in parent_component_inputs.artifacts, \
                'component_input_artifact: {} not found. All inputs: {}'.format(
                    component_input_artifact, parent_component_inputs)

                task_spec.inputs.artifacts[
                    input_name].component_input_artifact = component_input_artifact


def pop_input_from_component_spec(
    component_spec: pipeline_spec_pb2.ComponentSpec,
    input_name: str,
) -> None:
    """Removes an input from component spec input_definitions.

    Args:
      component_spec: The component spec to update in place.
      input_name: The name of the input, which could be an artifact or paremeter.
    """
    component_spec.input_definitions.artifacts.pop(input_name)
    component_spec.input_definitions.parameters.pop(input_name)

    if component_spec.input_definitions == pipeline_spec_pb2.ComponentInputsSpec(
    ):
        component_spec.ClearField('input_definitions')


def pop_input_from_task_spec(
    task_spec: pipeline_spec_pb2.PipelineTaskSpec,
    input_name: str,
) -> None:
    """Removes an input from task spec inputs.

    Args:
      task_spec: The pipeline task spec to update in place.
      input_name: The name of the input, which could be an artifact or paremeter.
    """
    task_spec.inputs.artifacts.pop(input_name)
    task_spec.inputs.parameters.pop(input_name)

    if task_spec.inputs == pipeline_spec_pb2.TaskInputsSpec():
        task_spec.ClearField('inputs')



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for `components` command group in KFP CLI."""
import contextlib
import pathlib
import sys
import textwrap
import unittest
from typing import List, Optional, Union
from unittest import mock

from typer import testing

# Docker is an optional install, but we need the import to succeed for tests.
# So we patch it before importing kfp.cli.components.
try:
    import docker  # pylint: disable=unused-import
except ImportError:
    sys.modules['docker'] = mock.Mock()
from kfp.deprecated.cli import components

_COMPONENT_TEMPLATE = '''
from kfp.dsl import *

@component(
  base_image={base_image},
  target_image={target_image},
  output_component_file={output_component_file})
def {func_name}():
    pass
'''


def _make_component(func_name: str,
                    base_image: Optional[str] = None,
                    target_image: Optional[str] = None,
                    output_component_file: Optional[str] = None) -> str:
    return textwrap.dedent('''
    from kfp.dsl import *

    @component(
        base_image={base_image},
        target_image={target_image},
        output_component_file={output_component_file})
    def {func_name}():
        pass
    ''').format(
        base_image=repr(base_image),
        target_image=repr(target_image),
        output_component_file=repr(output_component_file),
        func_name=func_name)


def _write_file(filename: str, file_contents: str):
    filepath = pathlib.Path(filename)
    filepath.parent.mkdir(exist_ok=True, parents=True)
    filepath.write_text(file_contents)


def _write_components(filename: str, component_template: Union[List[str], str]):
    if isinstance(component_template, list):
        file_contents = '\n\n'.join(component_template)
    else:
        file_contents = component_template
    _write_file(filename=filename, file_contents=file_contents)


class Test(unittest.TestCase):

    def setUp(self) -> None:
        self._runner = testing.CliRunner()
        components._DOCKER_IS_PRESENT = True

        patcher = mock.patch('docker.from_env')
        self._docker_client = patcher.start().return_value
        self._docker_client.images.build.return_value = [{
            'stream': 'Build logs'
        }]
        self._docker_client.images.push.return_value = [{'status': 'Pushed'}]
        self.addCleanup(patcher.stop)

        self._app = components.app
        with contextlib.ExitStack() as stack:
            stack.enter_context(self._runner.isolated_filesystem())
            self._working_dir = pathlib.Path.cwd()
            self.addCleanup(stack.pop_all().close)

        return super().setUp()

    def assertFileExists(self, path: str):
        path_under_test_dir = self._working_dir / path
        self.assertTrue(path_under_test_dir, f'File {path} does not exist!')

    def assertFileExistsAndContains(self, path: str, expected_content: str):
        self.assertFileExists(path)
        path_under_test_dir = self._working_dir / path
        got_content = path_under_test_dir.read_text()
        self.assertEqual(got_content, expected_content)

    def testKFPConfigForSingleFile(self):
        preprocess_component = _make_component(
            func_name='preprocess', target_image='custom-image')
        train_component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py',
                          [preprocess_component, train_component])

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)

        self.assertFileExistsAndContains(
            'kfp_config.ini',
            textwrap.dedent('''\
            [Components]
            preprocess = components.py
            train = components.py

            '''))

    def testKFPConfigForSingleFileUnderNestedDirectory(self):
        preprocess_component = _make_component(
            func_name='preprocess', target_image='custom-image')
        train_component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('dir1/dir2/dir3/components.py',
                          [preprocess_component, train_component])

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)

        self.assertFileExistsAndContains(
            'kfp_config.ini',
            textwrap.dedent('''\
            [Components]
            preprocess = dir1/dir2/dir3/components.py
            train = dir1/dir2/dir3/components.py

            '''))

    def testKFPConfigForMultipleFiles(self):
        component = _make_component(
            func_name='preprocess', target_image='custom-image')
        _write_components('preprocess_component.py', component)

        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('train_component.py', component)

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)

        self.assertFileExistsAndContains(
            'kfp_config.ini',
            textwrap.dedent('''\
            [Components]
            preprocess = preprocess_component.py
            train = train_component.py

            '''))

    def testKFPConfigForMultipleFilesUnderNestedDirectories(self):
        component = _make_component(
            func_name='preprocess', target_image='custom-image')
        _write_components('preprocess/preprocess_component.py', component)

        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('train/train_component.py', component)

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)

        self.assertFileExistsAndContains(
            'kfp_config.ini',
            textwrap.dedent('''\
            [Components]
            preprocess = preprocess/preprocess_component.py
            train = train/train_component.py

            '''))

    def testTargetImageMustBeTheSameInAllComponents(self):
        component_one = _make_component(func_name='one', target_image='image-1')
        component_two = _make_component(func_name='two', target_image='image-1')
        _write_components('one_two/one_two.py', [component_one, component_two])

        component_three = _make_component(
            func_name='three', target_image='image-2')
        component_four = _make_component(
            func_name='four', target_image='image-3')
        _write_components('three_four/three_four.py',
                          [component_three, component_four])

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 1)

    def testTargetImageMustBeTheSameInAllComponents(self):
        component_one = _make_component(
            func_name='one', base_image='image-1', target_image='target-image')
        component_two = _make_component(
            func_name='two', base_image='image-1', target_image='target-image')
        _write_components('one_two/one_two.py', [component_one, component_two])

        component_three = _make_component(
            func_name='three',
            base_image='image-2',
            target_image='target-image')
        component_four = _make_component(
            func_name='four', base_image='image-3', target_image='target-image')
        _write_components('three_four/three_four.py',
                          [component_three, component_four])

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 1)

    def testComponentFilepatternCanBeUsedToRestrictDiscovery(self):
        component = _make_component(
            func_name='preprocess', target_image='custom-image')
        _write_components('preprocess/preprocess_component.py', component)

        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('train/train_component.py', component)

        result = self._runner.invoke(
            self._app,
            [
                'build',
                str(self._working_dir), '--component-filepattern=train/*'
            ],
        )
        self.assertEqual(result.exit_code, 0)

        self.assertFileExistsAndContains(
            'kfp_config.ini',
            textwrap.dedent('''\
            [Components]
            train = train/train_component.py

            '''))

    def testEmptyRequirementsTxtFileIsGenerated(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self._runner.invoke(self._app,
                                     ['build', str(self._working_dir)])
        self.assertEqual(result.exit_code, 0)
        self.assertFileExistsAndContains('requirements.txt',
                                         '# Generated by KFP.\n')

    def testExistingRequirementsTxtFileIsUnchanged(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        _write_file('requirements.txt', 'Some pre-existing content')

        result = self._runner.invoke(self._app,
                                     ['build', str(self._working_dir)])
        self.assertEqual(result.exit_code, 0)
        self.assertFileExistsAndContains('requirements.txt',
                                         'Some pre-existing content')

    def testDockerignoreFileIsGenerated(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self._runner.invoke(self._app,
                                     ['build', str(self._working_dir)])
        self.assertEqual(result.exit_code, 0)
        self.assertFileExistsAndContains(
            '.dockerignore',
            textwrap.dedent('''\
            # Generated by KFP.

            component_metadata/
            '''))

    def testExistingDockerignoreFileIsUnchanged(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        _write_file('.dockerignore', 'Some pre-existing content')

        result = self._runner.invoke(self._app,
                                     ['build', str(self._working_dir)])
        self.assertEqual(result.exit_code, 0)
        self.assertFileExistsAndContains('.dockerignore',
                                         'Some pre-existing content')

    def testDockerEngineIsSupported(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir), '--engine=docker'])
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self._docker_client.images.push.assert_called_once_with(
            'custom-image', stream=True, decode=True)

    def testKanikoEngineIsNotSupported(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)
        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir), '--engine=kaniko'],
        )
        self.assertEqual(result.exit_code, 1)
        self._docker_client.api.build.assert_not_called()
        self._docker_client.images.push.assert_not_called()

    def testCloudBuildEngineIsNotSupported(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)
        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir), '--engine=cloudbuild'],
        )
        self.assertEqual(result.exit_code, 1)
        self._docker_client.api.build.assert_not_called()
        self._docker_client.images.push.assert_not_called()

    def testDockerClientIsCalledToBuildAndPushByDefault(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)

        self._docker_client.api.build.assert_called_once()
        self._docker_client.images.push.assert_called_once_with(
            'custom-image', stream=True, decode=True)

    def testDockerClientIsCalledToBuildButSkipsPushing(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir), '--no-push-image'],
        )
        self.assertEqual(result.exit_code, 0)

        self._docker_client.api.build.assert_called_once()
        self._docker_client.images.push.assert_not_called()

    @mock.patch('kfp.__version__', '1.2.3')
    def testDockerfileIsCreatedCorrectly(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self.assertFileExistsAndContains(
            'Dockerfile',
            textwrap.dedent('''\
                # Generated by KFP.

                FROM python:3.7

                WORKDIR /usr/local/src/kfp/components
                COPY requirements.txt requirements.txt
                RUN pip install --no-cache-dir -r requirements.txt

                RUN pip install --no-cache-dir kfp==1.8.11
                COPY . .
                '''))

    def testExistingDockerfileIsUnchangedByDefault(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)
        _write_file('Dockerfile', 'Existing Dockerfile contents')

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir)],
        )
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self.assertFileExistsAndContains('Dockerfile',
                                         'Existing Dockerfile contents')

    @mock.patch('kfp.__version__', '1.2.3')
    def testExistingDockerfileCanBeOverwritten(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)
        _write_file('Dockerfile', 'Existing Dockerfile contents')

        result = self._runner.invoke(
            self._app,
            ['build', str(self._working_dir), '--overwrite-dockerfile'],
        )
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self.assertFileExistsAndContains(
            'Dockerfile',
            textwrap.dedent('''\
                # Generated by KFP.

                FROM python:3.7

                WORKDIR /usr/local/src/kfp/components
                COPY requirements.txt requirements.txt
                RUN pip install --no-cache-dir -r requirements.txt

                RUN pip install --no-cache-dir kfp==1.8.11
                COPY . .
                '''))

    def testDockerfileCanContainCustomKFPPackage(self):
        component = _make_component(
            func_name='train', target_image='custom-image')
        _write_components('components.py', component)

        result = self._runner.invoke(
            self._app,
            [
                'build',
                str(self._working_dir),
                '--kfp-package-path=/Some/localdir/containing/kfp/source'
            ],
        )
        self.assertEqual(result.exit_code, 0)
        self._docker_client.api.build.assert_called_once()
        self.assertFileExistsAndContains(
            'Dockerfile',
            textwrap.dedent('''\
                # Generated by KFP.

                FROM python:3.7

                WORKDIR /usr/local/src/kfp/components
                COPY requirements.txt requirements.txt
                RUN pip install --no-cache-dir -r requirements.txt

                RUN pip install --no-cache-dir /Some/localdir/containing/kfp/source
                COPY . .
                '''))


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import logging
import os
import re
import subprocess
import tempfile
import warnings
from collections import deque
from typing import Any, Callable, Dict, List, Mapping, Optional, Union, cast

from . import dsl
from .compiler.compiler import sanitize_k8s_name


class _Dag:
    """DAG stands for Direct Acyclic Graph.

    DAG here is used to decide the order to execute pipeline ops.

    For more information on DAG, please refer to `wiki <https://en.wikipedia.org/wiki/Directed_acyclic_graph>`_.
    """

    def __init__(self, nodes: List[str]) -> None:
        """

        Args::
          nodes: List of DAG nodes, each node is identified by an unique name.
        """
        self._graph = {node: [] for node in nodes}
        self._reverse_graph = {node: [] for node in nodes}

    @property
    def graph(self):
        return self._graph

    @property
    def reverse_graph(self):
        return self._reverse_graph

    def add_edge(self, edge_source: str, edge_target: str) -> None:
        """Add an edge between DAG nodes.

        Args::
          edge_source: the source node of the edge
          edge_target: the target node of the edge
        """
        self._graph[edge_source].append(edge_target)
        self._reverse_graph[edge_target].append(edge_source)

    def get_follows(self, source_node: str) -> List[str]:
        """Get all target nodes start from the specified source node.

        Args::
          source_node: the source node
        """
        return self._graph.get(source_node, [])

    def get_dependencies(self, target_node: str) -> List[str]:
        """Get all source nodes end with the specified target node.

        Args::
          target_node: the target node
        """
        return self._reverse_graph.get(target_node, [])

    def topological_sort(self) -> List[str]:
        """List DAG nodes in topological order."""

        in_degree = {node: 0 for node in self._graph.keys()}

        for i in self._graph:
            for j in self._graph[i]:
                in_degree[j] += 1

        queue = deque()
        for node, degree in in_degree.items():
            if degree == 0:
                queue.append(node)

        sorted_nodes = []

        while queue:
            u = queue.popleft()
            sorted_nodes.append(u)

            for node in self._graph[u]:
                in_degree[node] -= 1

                if in_degree[node] == 0:
                    queue.append(node)

        return sorted_nodes


def _extract_pipeline_param(param: str) -> dsl.PipelineParam:
    """Extract PipelineParam from string."""
    matches = re.findall(r"{{pipelineparam:op=([\w\s_-]*);name=([\w\s_-]+)}}",
                         param)
    op_dependency_name = matches[0][0]
    output_file_name = matches[0][1]
    return dsl.PipelineParam(output_file_name, op_dependency_name)


def _get_op(ops: List[dsl.ContainerOp],
            op_name: str) -> Union[dsl.ContainerOp, None]:
    """Get the first op with specified op name."""
    return next(filter(lambda op: op.name == op_name, ops), None)


def _get_subgroup(groups: List[dsl.OpsGroup],
                  group_name: str) -> Union[dsl.OpsGroup, None]:
    """Get the first OpsGroup with specified group name."""
    return next(filter(lambda g: g.name == group_name, groups), None)


class LocalClient:

    class ExecutionMode:
        """Configuration to decide whether the client executes a component in
        docker or in local process."""

        DOCKER = "docker"
        LOCAL = "local"

        def __init__(
            self,
            mode: str = DOCKER,
            images_to_exclude: List[str] = [],
            ops_to_exclude: List[str] = [],
            docker_options: List[str] = [],
        ) -> None:
            """Constructor.

            Args:
                mode: Default execution mode, default 'docker'
                images_to_exclude: If the image of op is in images_to_exclude, the op is
                    executed in the mode different from default_mode.
                ops_to_exclude: If the name of op is in ops_to_exclude, the op is
                    executed in the mode different from default_mode.
                docker_options: Docker options used in docker mode,
                    e.g. docker_options=["-e", "foo=bar"].
            """
            if mode not in [self.DOCKER, self.LOCAL]:
                raise Exception(
                    "Invalid execution mode, must be docker of local")
            self._mode = mode
            self._images_to_exclude = images_to_exclude
            self._ops_to_exclude = ops_to_exclude
            self._docker_options = docker_options

        @property
        def mode(self) -> str:
            return self._mode

        @property
        def images_to_exclude(self) -> List[str]:
            return self._images_to_exclude

        @property
        def ops_to_exclude(self) -> List[str]:
            return self._ops_to_exclude

        @property
        def docker_options(self) -> List[str]:
            return self._docker_options

    def __init__(self, pipeline_root: Optional[str] = None) -> None:
        """Construct the instance of LocalClient.

        Args：
            pipeline_root: The root directory where the output artifact of component
              will be saved.
        """
        warnings.warn(
            'LocalClient is an Alpha[1] feature. It may be deprecated in the future.\n'
            '[1] https://github.com/kubeflow/pipelines/blob/master/docs/release/feature-stages.md#alpha',
            category=FutureWarning,
        )

        pipeline_root = pipeline_root or tempfile.tempdir
        self._pipeline_root = pipeline_root

    def _find_base_group(self, groups: List[dsl.OpsGroup],
                         op_name: str) -> Union[dsl.OpsGroup, None]:
        """Find the base group of op in candidate group list."""
        if groups is None or len(groups) == 0:
            return None
        for group in groups:
            if _get_op(group.ops, op_name):
                return group
            else:
                _parent_group = self._find_base_group(group.groups, op_name)
                if _parent_group:
                    return group

        return None

    def _create_group_dag(self, pipeline_dag: _Dag,
                          group: dsl.OpsGroup) -> _Dag:
        """Create DAG within current group, it's a DAG of direct ops and direct
        subgroups.

        Each node of the DAG is either an op or a subgroup. For each
        node in current group, if one of its DAG follows is also an op
        in current group, add an edge to this follow op, otherwise, if
        this follow belongs to subgroups, add an edge to its subgroup.
        If this node has dependency from subgroups, then add an edge
        from this subgroup to current node.
        """
        group_dag = _Dag([op.name for op in group.ops] +
                         [g.name for g in group.groups])

        for op in group.ops:
            for follow in pipeline_dag.get_follows(op.name):
                if _get_op(group.ops, follow) is not None:
                    # add edge between direct ops
                    group_dag.add_edge(op.name, follow)
                else:
                    _base_group = self._find_base_group(group.groups, follow)
                    if _base_group:
                        # add edge to direct subgroup
                        group_dag.add_edge(op.name, _base_group.name)

            for dependency in pipeline_dag.get_dependencies(op.name):
                if _get_op(group.ops, dependency) is None:
                    _base_group = self._find_base_group(group.groups,
                                                        dependency)
                    if _base_group:
                        # add edge from direct subgroup
                        group_dag.add_edge(_base_group.name, op.name)

        return group_dag

    def _create_op_dag(self, p: dsl.Pipeline) -> _Dag:
        """Create the DAG of the pipeline ops."""
        dag = _Dag(p.ops.keys())

        for op in p.ops.values():
            # dependencies defined by inputs
            for input_value in op.inputs:
                if isinstance(input_value, dsl.PipelineParam):
                    input_param = _extract_pipeline_param(input_value.pattern)
                    if input_param.op_name:
                        dag.add_edge(input_param.op_name, op.name)
                    else:
                        logging.debug("%s depend on pipeline param", op.name)

            # explicit dependencies of current op
            for dependent in op.dependent_names:
                dag.add_edge(dependent, op.name)
        return dag

    def _make_output_file_path_unique(self, run_name: str, op_name: str,
                                      output_file: str) -> str:
        """Alter the file path of output artifact to make sure it's unique in
        local runner.

        kfp compiler will bound a tmp file for each component output,
        which is unique in kfp runtime, but not unique in local runner.
        We alter the file path of the name of current run and op, to
        make it unique in local runner.
        """
        if not output_file.startswith("/tmp/"):
            return output_file
        return f'{self._pipeline_root}/{run_name}/{op_name.lower()}/{output_file[len("/tmp/"):]}'

    def _get_output_file_path(
        self,
        run_name: str,
        pipeline: dsl.Pipeline,
        op_name: str,
        output_name: str = None,
    ) -> str:
        """Get the file path of component output."""

        op_dependency = pipeline.ops[op_name]
        if output_name is None and len(op_dependency.file_outputs) == 1:
            output_name = next(iter(op_dependency.file_outputs.keys()))
        output_file = op_dependency.file_outputs[output_name]
        unique_output_file = self._make_output_file_path_unique(
            run_name, op_name, output_file)
        return unique_output_file

    def _generate_cmd_for_subprocess_execution(
        self,
        run_name: str,
        pipeline: dsl.Pipeline,
        op: dsl.ContainerOp,
        stack: Dict[str, Any],
    ) -> List[str]:
        """Generate shell command to run the op locally."""
        cmd = op.command + op.arguments

        # In debug mode, for `python -c cmd` format command, pydev will insert code before
        # `cmd`, but there is no newline at the end of the inserted code, which will cause
        # syntax error, so we add newline before `cmd`.
        for i in range(len(cmd)):
            if cmd[i] == "-c":
                cmd[i + 1] = "\n" + cmd[i + 1]

        for index, cmd_item in enumerate(cmd):
            if cmd_item in stack:  # Argument is LoopArguments item
                cmd[index] = str(stack[cmd_item])
            elif cmd_item in op.file_outputs.values(
            ):  # Argument is output file
                output_name = next(
                    filter(lambda item: item[1] == cmd_item,
                           op.file_outputs.items()))[0]
                output_param = op.outputs[output_name]
                output_file = cmd_item
                output_file = self._make_output_file_path_unique(
                    run_name, output_param.op_name, output_file)

                os.makedirs(os.path.dirname(output_file), exist_ok=True)
                cmd[index] = output_file
            elif (cmd_item in op.input_artifact_paths.values()
                 ):  # Argument is input artifact file
                input_name = next(
                    filter(
                        lambda item: item[1] == cmd_item,
                        op.input_artifact_paths.items(),
                    ))[0]
                input_param_pattern = op.artifact_arguments[input_name]
                pipeline_param = _extract_pipeline_param(input_param_pattern)
                input_file = self._get_output_file_path(run_name, pipeline,
                                                        pipeline_param.op_name,
                                                        pipeline_param.name)

                cmd[index] = input_file

        return cmd

    def _generate_cmd_for_docker_execution(
            self,
            run_name: str,
            pipeline: dsl.Pipeline,
            op: dsl.ContainerOp,
            stack: Dict[str, Any],
            docker_options: List[str] = []) -> List[str]:
        """Generate the command to run the op in docker locally."""
        cmd = self._generate_cmd_for_subprocess_execution(
            run_name, pipeline, op, stack)

        docker_cmd = [
            "docker",
            "run",
            *docker_options,
            "-v",
            "{pipeline_root}:{pipeline_root}".format(
                pipeline_root=self._pipeline_root),
            op.image,
        ] + cmd
        return docker_cmd

    def _run_group_dag(
        self,
        run_name: str,
        pipeline: dsl.Pipeline,
        pipeline_dag: _Dag,
        current_group: dsl.OpsGroup,
        stack: Dict[str, Any],
        execution_mode: ExecutionMode,
    ) -> bool:
        """Run ops in current group in topological order.

        Args:
            pipeline: kfp.dsl.Pipeline
            pipeline_dag: DAG of pipeline ops
            current_group: current ops group
            stack: stack to trace `LoopArguments`
            execution_mode: Configuration to decide whether the client executes
                component in docker or in local process.
        Returns:
            True if succeed to run the group dag.
        """
        group_dag = self._create_group_dag(pipeline_dag, current_group)

        for node in group_dag.topological_sort():
            subgroup = _get_subgroup(current_group.groups, node)
            if subgroup is not None:  # Node of DAG is subgroup
                success = self._run_group(run_name, pipeline, pipeline_dag,
                                          subgroup, stack, execution_mode)
                if not success:
                    return False
            else:  # Node of DAG is op
                op = _get_op(current_group.ops, node)

                execution_mode = (
                    execution_mode
                    if execution_mode else LocalClient.ExecutionMode())
                can_run_locally = execution_mode.mode == LocalClient.ExecutionMode.LOCAL
                exclude = (
                    op.image in execution_mode.images_to_exclude or
                    op.name in execution_mode.ops_to_exclude)
                if exclude:
                    can_run_locally = not can_run_locally

                if can_run_locally:
                    cmd = self._generate_cmd_for_subprocess_execution(
                        run_name, pipeline, op, stack)
                else:
                    cmd = self._generate_cmd_for_docker_execution(
                        run_name, pipeline, op, stack,
                        execution_mode.docker_options)
                process = subprocess.Popen(
                    cmd,
                    shell=False,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    universal_newlines=True,
                )
                # TODO support async process
                logging.info("start task：%s", op.name)
                stdout, stderr = process.communicate()
                if stdout:
                    logging.info(stdout)
                if stderr:
                    logging.error(stderr)
                if process.returncode != 0:
                    logging.error(cmd)
                    return False

    def _run_group(
        self,
        run_name: str,
        pipeline: dsl.Pipeline,
        pipeline_dag: _Dag,
        current_group: dsl.OpsGroup,
        stack: Dict[str, Any],
        execution_mode: ExecutionMode,
    ) -> bool:
        """Run all ops in current group.

        Args:
            run_name: str, the name of this run, can be used to query the run result
            pipeline: kfp.dsl.Pipeline
            pipeline_dag: DAG of pipeline ops
            current_group: current ops group
            stack: stack to trace `LoopArguments`
            execution_mode: Configuration to decide whether the client executes
                component in docker or in local process.
        Returns:
            True if succeed to run the group.
        """
        if current_group.type == dsl.ParallelFor.TYPE_NAME:
            current_group = cast(dsl.ParallelFor, current_group)

            if current_group.items_is_pipeline_param:
                _loop_args = current_group.loop_args
                _param_name = _loop_args.name[:-len(_loop_args
                                                    .LOOP_ITEM_NAME_BASE) - 1]

                _op_dependency = pipeline.ops[_loop_args.op_name]
                _list_file = _op_dependency.file_outputs[_param_name]
                _altered_list_file = self._make_output_file_path_unique(
                    run_name, _loop_args.op_name, _list_file)
                with open(_altered_list_file, "r") as f:
                    _param_values = json.load(f)
                for index, _param_value in enumerate(_param_values):
                    if isinstance(_param_values, (dict, list)):
                        _param_value = json.dumps(_param_value)
                    stack[_loop_args.pattern] = _param_value
                    loop_run_name = "{run_name}/{loop_index}".format(
                        run_name=run_name, loop_index=index)
                    success = self._run_group_dag(
                        loop_run_name,
                        pipeline,
                        pipeline_dag,
                        current_group,
                        stack,
                        execution_mode,
                    )
                    del stack[_loop_args.pattern]
                    if not success:
                        return False
                return True
            else:
                raise Exception("Not implemented")
        else:
            return self._run_group_dag(run_name, pipeline, pipeline_dag,
                                       current_group, stack, execution_mode)

    def create_run_from_pipeline_func(
            self,
            pipeline_func: Callable,
            arguments: Mapping[str, str],
            execution_mode: ExecutionMode = ExecutionMode(),
    ):
        """Runs a pipeline locally, either using Docker or in a local process.

        Parameters:
          pipeline_func: pipeline function
          arguments: Arguments to the pipeline function provided as a dict, reference
              to `kfp.client.create_run_from_pipeline_func`
          execution_mode: Configuration to decide whether the client executes component
              in docker or in local process.
        """

        class RunPipelineResult:

            def __init__(self, client: LocalClient, pipeline: dsl.Pipeline,
                         run_id: str, success: bool):
                self._client = client
                self._pipeline = pipeline
                self.run_id = run_id
                self._success = success

            def get_output_file(self, op_name: str, output: str = None):
                return self._client._get_output_file_path(
                    self.run_id, self._pipeline, op_name, output)

            def success(self) -> bool:
                return self._success

            def __repr__(self):
                return "RunPipelineResult(run_id={})".format(self.run_id)

        pipeline_name = sanitize_k8s_name(
            getattr(pipeline_func, "_component_human_name", None) or
            pipeline_func.__name__)
        with dsl.Pipeline(pipeline_name) as pipeline:
            pipeline_func(**arguments)

        run_version = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        run_name = pipeline.name.replace(" ", "_").lower() + "_" + run_version

        pipeline_dag = self._create_op_dag(pipeline)
        success = self._run_group(run_name, pipeline, pipeline_dag,
                                  pipeline.groups[0], {}, execution_mode)

        return RunPipelineResult(self, pipeline, run_name, success=success)



# Copyright 2020-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import tempfile
from typing import Any, Dict, List, Union
import unittest

from absl.testing import parameterized
import kfp
from kfp import compiler
from kfp import components
from kfp import dsl
from kfp.components import base_component
from kfp.components import pipeline_channel
from kfp.components import structures
from kfp.components import yaml_component
from kfp.components.types import artifact_types
from kfp.components.types import type_utils
from kfp.components.types.type_utils import InconsistentTypeException
from kfp.dsl import Dataset
from kfp.dsl import Input
from kfp.dsl import Output
from kfp.pipeline_spec import pipeline_spec_pb2 as pb

_PARAMETER_TYPES = [
    'String',
    'str',
    'Integer',
    'int',
    'Float',
    'Double',
    'bool',
    'Boolean',
    'Dict',
    'List',
    'JsonObject',
    'JsonArray',
    {
        'JsonObject': {
            'data_type': 'proto:tfx.components.trainer.TrainArgs'
        }
    },
    'PipelineTaskFinalStatus',
]

_KNOWN_ARTIFACT_TYPES = [
    'Model',
    'Dataset',
    'Schema',
    'Metrics',
    'ClassificationMetrics',
    'SlicedClassificationMetrics',
    'HTML',
    'Markdown',
]

_UNKNOWN_ARTIFACT_TYPES = [None, 'Arbtrary Model', 'dummy']


class _ArbitraryClass:
    pass


class _VertexDummy(artifact_types.Artifact):
    schema_title = 'google.VertexDummy'
    schema_version = '0.0.2'

    def __init__(self):
        super().__init__(uri='uri', name='name', metadata={'dummy': '123'})


class TypeUtilsTest(parameterized.TestCase):

    @parameterized.parameters(
        [(item, True) for item in _PARAMETER_TYPES] +
        [(item, False)
         for item in _KNOWN_ARTIFACT_TYPES + _UNKNOWN_ARTIFACT_TYPES])
    def test_is_parameter_type_true(self, type_name, expected_result):
        self.assertEqual(expected_result,
                         type_utils.is_parameter_type(type_name))

    @parameterized.parameters(
        {
            'given_type': 'Int',
            'expected_type': pb.ParameterType.NUMBER_INTEGER,
        },
        {
            'given_type': 'Integer',
            'expected_type': pb.ParameterType.NUMBER_INTEGER,
        },
        {
            'given_type': int,
            'expected_type': pb.ParameterType.NUMBER_INTEGER,
        },
        {
            'given_type': 'Double',
            'expected_type': pb.ParameterType.NUMBER_DOUBLE,
        },
        {
            'given_type': 'Float',
            'expected_type': pb.ParameterType.NUMBER_DOUBLE,
        },
        {
            'given_type': float,
            'expected_type': pb.ParameterType.NUMBER_DOUBLE,
        },
        {
            'given_type': 'String',
            'expected_type': pb.ParameterType.STRING,
        },
        {
            'given_type': 'Text',
            'expected_type': pb.ParameterType.STRING,
        },
        {
            'given_type': str,
            'expected_type': pb.ParameterType.STRING,
        },
        {
            'given_type': 'Boolean',
            'expected_type': pb.ParameterType.BOOLEAN,
        },
        {
            'given_type': bool,
            'expected_type': pb.ParameterType.BOOLEAN,
        },
        {
            'given_type': 'Dict',
            'expected_type': pb.ParameterType.STRUCT,
        },
        {
            'given_type': dict,
            'expected_type': pb.ParameterType.STRUCT,
        },
        {
            'given_type': 'List',
            'expected_type': pb.ParameterType.LIST,
        },
        {
            'given_type': list,
            'expected_type': pb.ParameterType.LIST,
        },
        {
            'given_type': Dict[str, int],
            'expected_type': pb.ParameterType.STRUCT,
        },
        {
            'given_type': List[Any],
            'expected_type': pb.ParameterType.LIST,
        },
        {
            'given_type': {
                'JsonObject': {
                    'data_type': 'proto:tfx.components.trainer.TrainArgs'
                }
            },
            'expected_type': pb.ParameterType.STRUCT,
        },
    )
    def test_get_parameter_type(self, given_type, expected_type):
        self.assertEqual(expected_type,
                         type_utils.get_parameter_type(given_type))

        # Test get parameter by Python type.
        self.assertEqual(pb.ParameterType.NUMBER_INTEGER,
                         type_utils.get_parameter_type(int))

    def test_get_parameter_type_invalid(self):
        with self.assertRaises(AttributeError):
            type_utils.get_parameter_type_schema(None)

    @parameterized.parameters(
        {
            'given_type': str,
            'expected_type_name': 'String',
        },
        {
            'given_type': int,
            'expected_type_name': 'Integer',
        },
        {
            'given_type': float,
            'expected_type_name': 'Float',
        },
        {
            'given_type': bool,
            'expected_type_name': 'Boolean',
        },
        {
            'given_type': list,
            'expected_type_name': 'List',
        },
        {
            'given_type': dict,
            'expected_type_name': 'Dict',
        },
        {
            'given_type': Any,
            'expected_type_name': None,
        },
    )
    def test_get_canonical_type_name_for_type(
        self,
        given_type,
        expected_type_name,
    ):
        self.assertEqual(
            expected_type_name,
            type_utils.get_canonical_type_name_for_type(given_type))

    @parameterized.parameters(
        {
            'given_type': 'PipelineTaskFinalStatus',
            'expected_result': True,
        },
        {
            'given_type': 'pipelineTaskFinalstatus',
            'expected_result': False,
        },
        {
            'given_type': int,
            'expected_result': False,
        },
    )
    def test_is_task_final_statu_type(self, given_type, expected_result):
        self.assertEqual(expected_result,
                         type_utils.is_task_final_status_type(given_type))


class TestGetArtifactTypeSchema(parameterized.TestCase):

    @parameterized.parameters([
        # v2 standard system types
        {
            'schema_title': 'system.Artifact@0.0.1',
            'exp_schema_title': 'system.Artifact',
            'exp_schema_version': '0.0.1',
        },
        {
            'schema_title': 'system.Dataset@0.0.1',
            'exp_schema_title': 'system.Dataset',
            'exp_schema_version': '0.0.1',
        },
        # google type with schema_version
        {
            'schema_title': 'google.VertexDataset@0.0.2',
            'exp_schema_title': 'google.VertexDataset',
            'exp_schema_version': '0.0.2',
        },
    ])
    def test_valid(
        self,
        schema_title: str,
        exp_schema_title: str,
        exp_schema_version: str,
    ):
        artifact_type_schema = type_utils.bundled_artifact_to_artifact_proto(
            schema_title)
        self.assertEqual(artifact_type_schema.schema_title, exp_schema_title)
        self.assertEqual(artifact_type_schema.schema_version,
                         exp_schema_version)


class TestTypeCheckManager(unittest.TestCase):

    def check_false_to_true(self):
        kfp.TYPE_CHECK = False
        with type_utils.TypeCheckManager(enable=True):
            self.assertEqual(kfp.TYPE_CHECK, True)
        self.assertEqual(kfp.TYPE_CHECK, False)

    def test_true_to_false(self):
        kfp.TYPE_CHECK = True
        with type_utils.TypeCheckManager(enable=False):
            self.assertEqual(kfp.TYPE_CHECK, False)
        self.assertEqual(kfp.TYPE_CHECK, True)


class TestCreateBundledArtifacttType(parameterized.TestCase):

    @parameterized.parameters([
        {
            'schema_title': 'system.Artifact',
            'schema_version': '0.0.2',
            'expected': 'system.Artifact@0.0.2'
        },
        {
            'schema_title': 'google.Artifact',
            'schema_version': '0.0.3',
            'expected': 'google.Artifact@0.0.3'
        },
        {
            'schema_title': 'system.Artifact',
            'schema_version': None,
            'expected': 'system.Artifact@0.0.1'
        },
        {
            'schema_title': 'google.Artifact',
            'schema_version': None,
            'expected': 'google.Artifact@0.0.1'
        },
    ])
    def test(self, schema_title: str, schema_version: Union[str, None],
             expected: str):
        actual = type_utils.create_bundled_artifact_type(
            schema_title, schema_version)
        self.assertEqual(actual, expected)


class TestValidateBundledArtifactType(parameterized.TestCase):

    @parameterized.parameters([
        {
            'type_': 'system.Artifact@0.0.1'
        },
        {
            'type_': 'system.Dataset@2.0.1'
        },
        {
            'type_': 'google.Model@2.0.0'
        },
    ])
    def test_valid(self, type_: str):
        type_utils.validate_bundled_artifact_type(type_)

    @parameterized.parameters([
        {
            'type_': 'system.Artifact'
        },
        {
            'type_': '2.0.1'
        },
        {
            'type_': 'google.Model2.0.0'
        },
        {
            'type_': 'google.Model2.0.0'
        },
        {
            'type_': 'google.Model@'
        },
        {
            'type_': 'google.Model@'
        },
        {
            'type_': '@2.0.0'
        },
    ])
    def test_missing_part(self, type_: str):
        with self.assertRaisesRegex(
                TypeError,
                r'Artifacts must have both a schema_title and a schema_version, separated by `@`'
        ):
            type_utils.validate_bundled_artifact_type(type_)

    @parameterized.parameters([
        {
            'type_': 'system@0.0.1'
        },
        {
            'type_': 'google@0.0.1'
        },
        {
            'type_': 'other@0.0.1'
        },
        {
            'type_': 'Artifact@0.0.1'
        },
    ])
    def test_one_part_schema_title(self, type_: str):
        with self.assertRaisesRegex(
                TypeError,
                r'Artifact schema_title must have both a namespace and a name'):
            type_utils.validate_bundled_artifact_type(type_)

    @parameterized.parameters([
        {
            'type_': 'other.Artifact@0.0.1'
        },
    ])
    def test_must_be_system_or_google_namespace(self, type_: str):
        with self.assertRaisesRegex(
                TypeError,
                r'Artifact schema_title must belong to `system` or `google` namespace'
        ):
            type_utils.validate_bundled_artifact_type(type_)

    @parameterized.parameters([
        {
            'type_': 'system.Artifact@0'
        },
        {
            'type_': 'system.Artifact@0.0'
        },
        {
            'type_': 'google.Artifact@0.01'
        },
    ])
    def test_must_be_valid_semantic_version(self, type_: str):
        with self.assertRaisesRegex(
                TypeError,
                r'Artifact schema_version must use three-part semantic versioning'
        ):
            type_utils.validate_bundled_artifact_type(type_)


class TestDeserializeV1ComponentYamlDefault(parameterized.TestCase):

    @parameterized.parameters([
        {
            'type_': 'String',
            'default': 'val',
            'expected_type': str,
            'expected_val': 'val',
        },
        {
            'type_': 'Boolean',
            'default': 'True',
            'expected_type': bool,
            'expected_val': True,
        },
        {
            'type_': 'Boolean',
            'default': 'true',
            'expected_type': bool,
            'expected_val': True,
        },
        {
            'type_': 'Boolean',
            'default': 'False',
            'expected_type': bool,
            'expected_val': False,
        },
        {
            'type_': 'Boolean',
            'default': 'false',
            'expected_type': bool,
            'expected_val': False,
        },
        {
            'type_': 'Float',
            'default': '0.0',
            'expected_type': float,
            'expected_val': 0.0,
        },
        {
            'type_': 'Float',
            'default': '1.0',
            'expected_type': float,
            'expected_val': 1.0,
        },
        {
            'type_': 'Integer',
            'default': '0',
            'expected_type': int,
            'expected_val': 0,
        },
        {
            'type_': 'JsonObject',
            'default': '[]',
            'expected_type': list,
            'expected_val': [],
        },
        {
            'type_': 'JsonObject',
            'default': '[1, 1.0, "a", true]',
            'expected_type': list,
            'expected_val': [1, 1.0, 'a', True],
        },
        {
            'type_': 'JsonObject',
            'default': '{}',
            'expected_type': dict,
            'expected_val': {},
        },
        {
            'type_': 'JsonObject',
            'default': '{"a": 1.0, "b": true}',
            'expected_type': dict,
            'expected_val': {
                'a': 1.0,
                'b': True
            },
        },
    ])
    def test_for_defaults_as_strings(
        self,
        type_: Any,
        default: str,
        expected_type: type,
        expected_val: Any,
    ):
        res = type_utils.deserialize_v1_component_yaml_default(type_, default)
        # check type first since equals check is insufficient since 1.0 == 1
        self.assertIsInstance(res, expected_type)
        self.assertEquals(res, expected_val)

    @parameterized.parameters([
        {
            'type_': 'Boolean',
            'default': True,
            'expected_type': bool,
            'expected_val': True,
        },
        {
            'type_': 'Boolean',
            'default': False,
            'expected_type': bool,
            'expected_val': False,
        },
        {
            'type_': 'Float',
            'default': 0.0,
            'expected_type': float,
            'expected_val': 0.0,
        },
        {
            'type_': 'Float',
            'default': 1.0,
            'expected_type': float,
            'expected_val': 1.0,
        },
        {
            'type_': 'Integer',
            'default': 0,
            'expected_type': int,
            'expected_val': 0,
        },
        {
            'type_': 'JsonObject',
            'default': [],
            'expected_type': list,
            'expected_val': [],
        },
        {
            'type_': 'JsonObject',
            'default': [1, 1.0, 'a', True],
            'expected_type': list,
            'expected_val': [1, 1.0, 'a', True],
        },
        {
            'type_': 'JsonObject',
            'default': {},
            'expected_type': dict,
            'expected_val': {},
        },
        {
            'type_': 'JsonObject',
            'default': {
                'a': 1.0,
                'b': True
            },
            'expected_type': dict,
            'expected_val': {
                'a': 1.0,
                'b': True
            },
        },
    ])
    def test_robustness_to_literals(
        self,
        type_: Any,
        default: str,
        expected_type: type,
        expected_val: Any,
    ):
        res = type_utils.deserialize_v1_component_yaml_default(type_, default)
        # check type first since equals check is insufficient since 1.0 == 1
        self.assertIsInstance(res, expected_type)
        self.assertEquals(res, expected_val)


class TestTypeChecking(parameterized.TestCase):

    @parameterized.parameters(
        # param True
        {
            'argument_value': 'my text',
            'parameter_input_spec': structures.InputSpec(type='String'),
            'is_compatible': True,
        },
        # param False
        {
            'argument_value': 'my text',
            'parameter_input_spec': structures.InputSpec(type='Integer'),
            'is_compatible': False,
        },
        # param Artifact compat, irrespective of version
        {
            'argument_value':
                pipeline_channel.create_pipeline_channel(
                    'dummy_channel', 'system.Artifact@1.0.0'),
            'parameter_input_spec':
                structures.InputSpec('system.Model@0.0.1'),
            'is_compatible':
                True,
        },
        # param Artifact compat, irrespective of version, other way
        {
            'argument_value':
                pipeline_channel.create_pipeline_channel(
                    'dummy_channel', 'system.Metrics@1.0.0'),
            'parameter_input_spec':
                structures.InputSpec('system.Artifact@0.0.1'),
            'is_compatible':
                True,
        },
        # different schema_title incompat, irrespective of version
        {
            'argument_value':
                pipeline_channel.create_pipeline_channel(
                    'dummy_channel', 'system.Metrics@1.0.0'),
            'parameter_input_spec':
                structures.InputSpec('system.Dataset@1.0.0'),
            'is_compatible':
                False,
        },
        # different major version incompat
        {
            'argument_value':
                pipeline_channel.create_pipeline_channel(
                    'dummy_channel', 'system.Metrics@1.0.0'),
            'parameter_input_spec':
                structures.InputSpec('system.Metrics@2.1.1'),
            'is_compatible':
                False,
        },
        # namespace must match
        {
            'argument_value':
                pipeline_channel.create_pipeline_channel(
                    'dummy_channel', 'google.Model@1.0.0'),
            'parameter_input_spec':
                structures.InputSpec('system.Model@1.0.0'),
            'is_compatible':
                False,
        },
        # system.Artifact compatible works across namespace
        {
            'argument_value':
                pipeline_channel.create_pipeline_channel(
                    'dummy_channel', 'google.Model@1.0.0'),
            'parameter_input_spec':
                structures.InputSpec('system.Artifact@1.0.0'),
            'is_compatible':
                True,
        },
        {
            'argument_value':
                pipeline_channel.create_pipeline_channel(
                    'dummy_channel',
                    'system.Artifact@1.0.0',
                    is_artifact_list=True),
            'parameter_input_spec':
                structures.InputSpec(
                    'system.Artifact@1.0.0', is_artifact_list=True),
            'is_compatible':
                True,
        },
        {
            'argument_value':
                pipeline_channel.create_pipeline_channel(
                    'dummy_channel',
                    'system.Artifact@1.0.0',
                    is_artifact_list=True),
            'parameter_input_spec':
                structures.InputSpec(
                    'system.Artifact@1.0.0', is_artifact_list=False),
            'is_compatible':
                False,
        },
        {
            'argument_value':
                pipeline_channel.create_pipeline_channel(
                    'dummy_channel',
                    'system.Artifact@1.0.0',
                    is_artifact_list=False),
            'parameter_input_spec':
                structures.InputSpec(
                    'system.Artifact@1.0.0', is_artifact_list=True),
            'is_compatible':
                False,
        },
        {
            'argument_value':
                pipeline_channel.create_pipeline_channel(
                    'dummy_channel',
                    'system.Artifact@1.0.0',
                    is_artifact_list=True),
            'parameter_input_spec':
                structures.InputSpec('String'),
            'is_compatible':
                False,
        },
        {
            'argument_value':
                pipeline_channel.create_pipeline_channel(
                    'dummy_channel', 'String'),
            'parameter_input_spec':
                structures.InputSpec(
                    'system.Artifact@1.0.0', is_artifact_list=True),
            'is_compatible':
                False,
        },
    )
    def test_verify_type_compatibility(
        self,
        argument_value: Union[pipeline_channel.PipelineChannel, str, bool, int,
                              float, dict, list],
        parameter_input_spec: structures.InputSpec,
        is_compatible: bool,
    ):
        if is_compatible:
            self.assertTrue(
                type_utils.verify_type_compatibility(
                    given_value=argument_value,
                    expected_spec=parameter_input_spec,
                    error_message_prefix='',
                ))
        else:
            with self.assertRaises(InconsistentTypeException):
                type_utils.verify_type_compatibility(
                    given_value=argument_value,
                    expected_spec=parameter_input_spec,
                    error_message_prefix='',
                )

    def test_list_of_artifacts_across_compilation_valid(self):

        @dsl.component
        def double(num: int, out_dataset: Output[Dataset]):
            ...

        @dsl.component
        def add(in_datasets: Input[List[Dataset]],
                out_dataset: Output[Dataset]):
            ...

        loaded_add = compile_and_load_component(add)

        @dsl.pipeline
        def math_pipeline(in_datasets: Input[List[Dataset]]) -> List[Dataset]:
            # artifact list arg can be provided for artifact list type (of component)
            loaded_add(in_datasets=in_datasets)

            with dsl.ParallelFor([1, 2, 3]) as x:
                t = double(num=x)
            return dsl.Collected(t.outputs['out_dataset'])

        loaded_math_pipeline = compile_and_load_component(math_pipeline)

        @dsl.pipeline
        def outer_math_pipeline(
                in_datasets: Input[List[Dataset]]) -> List[Dataset]:
            # artifact list arg can be provided for artifact list type (of pipeline)
            t = loaded_math_pipeline(in_datasets=in_datasets)
            # artifact list output can be surfaced to artifact list output
            return t.output

    def test_list_of_artifacts_across_compilation_invalid_component_input(self):

        @dsl.component
        def add(in_datasets: Input[Dataset]):
            ...

        loaded_add = compile_and_load_component(add)

        with self.assertRaisesRegex(
                InconsistentTypeException,
                r"Incompatible argument passed to the input 'in_datasets' of component 'add': Argument type 'List\[system\.Dataset@0\.0\.1\]' is incompatible with the input type 'system\.Dataset@0\.0\.1'"
        ):

            @dsl.pipeline
            def math_pipeline(in_datasets: Input[List[Dataset]]):
                # artifact list arg cannot be provided to single artifact type (of component)
                loaded_add(in_datasets=in_datasets)

    def test_list_of_artifacts_across_compilation_invalid_pipeline_input(self):

        @dsl.component
        def double(num: int, out_dataset: Output[Dataset]):
            ...

        @dsl.component
        def add(in_datasets: Input[List[Dataset]],
                out_dataset: Output[Dataset]):
            ...

        @dsl.pipeline
        def math_pipeline(in_datasets: Input[List[Dataset]]) -> List[Dataset]:
            add(in_datasets=in_datasets)

            with dsl.ParallelFor([1, 2, 3]) as x:
                t = double(num=x)
            return dsl.Collected(t.outputs['out_dataset'])

        with self.assertRaisesRegex(
                InconsistentTypeException,
                r"Incompatible argument passed to the input 'in_datasets' of component 'math-pipeline': Argument type 'system\.Dataset@0\.0\.1' is incompatible with the input type 'List\[system.Dataset@0\.0\.1\]'"
        ):

            @dsl.pipeline
            def outer_math_pipeline(
                    in_dataset: Input[Dataset]) -> List[Dataset]:
                # single artifact arg cannot be provided to artifact list type (of pipeline)
                t = math_pipeline(in_datasets=in_dataset)
                return t.output

    def test_list_of_artifacts_across_compilation_invalid_pipeline_output(self):

        @dsl.component
        def double(num: int, out_dataset: Output[Dataset]):
            ...

        @dsl.component
        def add(in_datasets: Input[List[Dataset]],
                out_dataset: Output[Dataset]):
            ...

        @dsl.pipeline
        def math_pipeline(in_datasets: Input[List[Dataset]]) -> List[Dataset]:
            add(in_datasets=in_datasets)

            with dsl.ParallelFor([1, 2, 3]) as x:
                t = double(num=x)
            return dsl.Collected(t.outputs['out_dataset'])

        with self.assertRaisesRegex(
                InconsistentTypeException,
                r"Incompatible return type provided for output of pipeline 'outer-math-pipeline'\. Output of type 'List\[system\.Dataset@0\.0\.1\]' cannot be surfaced as pipeline output type 'system\.Dataset@0\.0\.1"
        ):

            @dsl.pipeline
            def outer_math_pipeline(
                    in_dataset: Input[List[Dataset]]) -> Dataset:
                t = math_pipeline(in_datasets=in_dataset)
                # artifact list output cannot be surfaced as single artifact output
                return t.output


def compile_and_load_component(
        comp: base_component.BaseComponent) -> yaml_component.YamlComponent:

    with tempfile.TemporaryDirectory() as tempdir:
        output_path = os.path.join(tempdir, 'pipeline.yaml')
        compiler.Compiler().compile(comp, output_path)
        return components.load_component_from_file(output_path)


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for `dsl compile` command group in KFP CLI."""
import unittest

from kfp import dsl
from kfp.cli import compile_


class TestIsPipelineOrComponent(unittest.TestCase):

    def test_is_pipeline(self):

        @dsl.component
        def my_comp():
            pass

        @dsl.container_component
        def my_container_comp():
            return dsl.ContainerSpec(
                image='python:3.7',
                command=['echo', 'hello world'],
                args=[],
            )

        def my_func(a: str, b: int):
            pass

        @dsl.pipeline(name='my-pipeline')
        def my_pipeline(a: str, b: int):
            my_comp()

        self.assertFalse(compile_.is_pipeline_func(my_comp))
        self.assertFalse(compile_.is_pipeline_func(my_container_comp))
        self.assertFalse(compile_.is_pipeline_func(my_func))
        self.assertTrue(compile_.is_pipeline_func(my_pipeline))

    def test_is_component(self):

        @dsl.component
        def my_comp():
            pass

        @dsl.container_component
        def my_container_comp():
            return dsl.ContainerSpec(
                image='python:3.7',
                command=['echo', 'hello world'],
                args=[],
            )

        def my_func(a: str, b: int):
            pass

        @dsl.pipeline(name='my-pipeline')
        def my_pipeline(a: str, b: int):
            my_comp()

        self.assertTrue(compile_.is_component_func(my_comp))
        self.assertTrue(compile_.is_component_func(my_container_comp))
        self.assertFalse(compile_.is_component_func(my_func))
        self.assertFalse(compile_.is_component_func(my_pipeline))


if __name__ == '__main__':
    unittest.main()
